"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import json
import time

from programy.utils.logging.ylogger import YLogger

from programy.clients.render.renderer import RichMediaRenderer


class FacebookRenderer(RichMediaRenderer):

    def __init__(self, client):
        RichMediaRenderer.__init__(self, client)

    def print_payload(self, title, payload, indent=2, sort_keys=False):
        print(title, json.dumps(payload, indent=indent, sort_keys=sort_keys))

    def send_payload(self, payload):
        result = self._client.facebook_bot.send_raw(payload)
        return result

    def handle_text(self, client_context, text):
        YLogger.debug(client_context, "Handling text...")

        payload = {
            'recipient': {
                'id': client_context.userid
            },
            'message': {
                'text': text['text']
            }
        }
        return self.send_payload(payload)

    def handle_url_button(self, client_context, button):
        YLogger.debug(client_context, "Handling url...")

        payload = {
            "recipient": {
                "id": client_context.userid
            },
            "message": {
                "attachment": {
                    "type": "template",
                    "payload": {
                        "template_type": "button",
                        "text": button['text'],
                        "buttons": [
                            {
                                "type": "web_url",
                                "url": button['url'],
                                "title": button['text']
                            }
                        ]
                    }
                }
            }

        }
        return self.send_payload(payload)

    def create_postback_button(self, userid, text, postback):
        return {
            "recipient": {
                "id": userid
            },
            "message": {
                "attachment": {
                    "type": "template",
                    "payload": {
                        "template_type": "button",
                        "text": text,
                        "buttons": [
                            {
                                "type": "postback",
                                "title": text,
                                "payload": postback
                            }
                        ]
                    }
                }
            }
        }

    def handle_postback_button(self, client_context, button):
        YLogger.debug(client_context, "Handling postback button...")

        payload = self.create_postback_button(client_context.userid, button['text'], button['postback'])
        return self.send_payload(payload)

    def handle_link(self, client_context, link):
        YLogger.debug(client_context, "Handling link...")

        payload = {
            "recipient": {
                "id": client_context.userid
            },
            "message": {
                "attachment": {
                    "type": "template",
                    "payload": {
                        "template_type": "button",
                        "text": link['text'],
                        "buttons": [
                            {
                                "type": "web_url",
                                "title": link['text'],
                                "url": link['url']
                            }
                        ]
                    }
                }
            }
        }
        return self.send_payload(payload)

    def handle_image(self, client_context, image):
        YLogger.debug(client_context, "Handling image...")

        payload = {
            "recipient": {
                "id": client_context.userid
            },
            "message": {
                "attachment": {
                    "type": "template",
                    "payload": {
                        "template_type": "media",
                        "elements": [
                            {
                                "media_type": "image",
                                "url": image['url']
                            }
                        ]
                    }
                }
            }
        }
        return self.send_payload(payload)

    def handle_video(self, client_context, video):
        YLogger.debug(client_context, "Handling video...")

        payload = {
            "recipient": {
                "id": client_context.userid
            },
            "message": {
                "attachment": {
                    "type": "template",
                    "payload": {
                        "template_type": "media",
                        "elements": [
                            {
                                "media_type": "video",
                                "url": video['url']
                            }
                        ]
                    }
                }
            }
        }
        return self.send_payload(payload)

    def create_card_payload(self, card):
        payload = {
            "title": card['title'],
            "image_url": card['image'],
            "subtitle": card['subtitle'],
            "buttons": []
        }

        for button in card['buttons']:
            if button['url'] is not None:
                payload['buttons'].append({
                    "type": "web_url",
                    "title": button['text'],
                    "url": button['url']
                })
            else:
                payload['buttons'].append({
                    "type": "web_url",
                    "title": button['text'],
                    "postback": button['postback']
                })

        return payload

    def handle_card(self, client_context, card):
        YLogger.debug(client_context, "Handling card...")

        if len(card['buttons']) > 3:
            YLogger.warning(client_context, "More buttons than facebook allows for a card")

        payload = {
            'recipient': {
                'id': client_context.userid
            },
            'message': {
                "attachment": {
                    "type": "template",
                    "payload": {
                        "template_type": "generic",
                        "elements": [
                        ]
                    }
                }
            }
        }
        payload['message']['attachment']['payload']['elements'].append(self.create_card_payload(card))
        return self.send_payload(payload)

    def handle_carousel(self, client_context, carousel):
        YLogger.debug(client_context, "Handling carousel...")

        if len(carousel['cards']) > 10:
            YLogger.warning(client_context, "More cards than facebook allows for a carousel")

        payload = {
            'recipient': {
                'id': client_context.userid
            },
            'message': {
                "attachment": {
                    "type": "template",
                    "payload": {
                        "template_type": "generic",
                        "elements": []
                    }
                }
            }
        }

        for card in carousel['cards']:

            element = {
                "title": card['title'],
                "image_url": card['image'],
                "subtitle": card['subtitle'],
                "buttons": []
            }

            for button in card['buttons']:
                if button['url'] is not None:
                    element['buttons'].append({
                        "type": "web_url",
                        "title": button['text'],
                        "url": button['url']
                    })
                else:
                    element['buttons'].append({
                        "type": "web_url",
                        "title": button['text'],
                        "postback": button['postback']
                    })

            payload['message']['attachment']['payload']['elements'].append(element)

        return self.send_payload(payload)

    def handle_reply(self, client_context, reply):
        YLogger.debug(client_context, "Handling reply...")

        if reply['postback'] is None:
            payload = self.create_postback_button(client_context.userid, reply['text'], reply['text'])
        else:
            payload = self.create_postback_button(client_context.userid, reply['text'], reply['postback'])
        self.send_payload(payload)

    def handle_delay(self, client_context, delay):
        YLogger.debug(client_context, "Handling delay...")
        payload = {
            "recipient": {
                "id": client_context.userid
            },
            "sender_action": "typing_on"
        }
        self.send_payload(payload)
        time.sleep(int(delay['seconds']))
        payload = {
            "recipient": {
                "id": client_context.userid
            },
            "sender_action": "typing_off"
        }
        return self.send_payload(payload)

    def handle_split(self, client_context, split):
        YLogger.debug(client_context, "Handling split...")

    def convert_to_element(self, item):
        if item["type"] == 'card':
            return self.create_card_payload(item)
        return None

    def handle_list(self, client_context, lst):
        YLogger.debug(client_context, "Handling list...")
        payload = {
            "recipient": {
                "id": client_context.userid
            },
            "message": {
                "attachment": {
                    "type": "template",
                    "payload": {
                        "template_type": "list",
                        "top_element_style": "compact",
                        "elements": [
                        ]
                    }
                }
            }
        }

        for item in lst.get('items', []):
            element = self.convert_to_element(item)
            if element is not None:
                payload["message"]["attachment"]["payload"]["elements"].append(element)

        return self.send_payload(payload)

    def handle_ordered_list(self, client_context, lst):
        YLogger.debug(client_context, "Handling ordered...")

        payload = {
            "recipient": {
                "id": client_context.userid
            },
            "message": {
                "attachment": {
                    "type": "template",
                    "payload": {
                        "template_type": "list",
                        "top_element_style": "compact",
                        "elements": [
                        ]
                    }
                }
            }
        }

        for item in lst.get('items', []):
            element = self.convert_to_element(item)
            if element is not None:
                payload["message"]["attachment"]["payload"]["elements"].append(element)

        return self.send_payload(payload)

    def handle_location(self, client_context, location):
        YLogger.debug(client_context, "Handling location...")

        payload = {
            'recipient': {
                'id': client_context.userid
            },
            "message": {
                "text": "Your location",
                "quick_replies": [
                    {
                        "content_type": "location"
                    }
                ]
            }
        }
        return self.send_payload(payload)



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from urllib.error import URLError
from textblob import TextBlob
from textblob.exceptions import NotTranslated
from textblob.exceptions import TranslatorError
from programy.utils.logging.ylogger import YLogger
from programy.nlp.translate.base import BaseTranslator

languages = {
    "AFRIKAANS": "AF",
    "ALBANIAN": "SQ",
    "AMHARIC": "AM",
    "ARABIC": "AR",
    "ARMENIAN": "HY",
    "AZEERBAIJANI": "AZ",
    "BASQUE": "EU",
    "BELARUSIAN": "BE",
    "BENGALI": "BN",
    "BOSNIAN": "BS",
    "BULGARIAN": "BG",
    "CATALAN": "CA",
    "CEBUANO": "CEB",
    "SIMPLIFIED CHINESE": "ZH-CN",
    "CHINESE": "ZH-TW",
    "TRADITIONAL CHINESE": "ZH-TW",
    "CORSICAN": "CO",
    "CROATIAN": "HR",
    "CZECH": "CS",
    "DANISH": "DA",
    "DUTCH": "NL",
    "ENGLISH": "EN",
    "ESPERANTO": "EO",
    "ESTONIAN": "ET",
    "FINNISH": "FI",
    "FRENCH": "FR",
    "FRISIAN": "FY",
    "GALICIAN": "GL",
    "GEORGIAN": "KA",
    "GERMAN": "DE",
    "GREEK": "EL",
    "GUJARATI": "GU",
    "HAITIAN": "HT",
    "CREOLE": "HT",
    "HAUSA": "HA",
    "HAWAIIAN": "HAW",
    "HEBREW": "HE**",
    "HINDI": "HI",
    "HMONG": "HMN",
    "HUNGARIAN": "HU",
    "ICELANDIC": "IS",
    "IGBO": "IG",
    "INDONESIAN": "ID",
    "IRISH": "GA",
    "ITALIAN": "IT",
    "JAPANESE": "JA",
    "JAVANESE": "JW",
    "KANNADA": "KN",
    "KAZAKH": "KK",
    "KHMER": "KM",
    "KOREAN": "KO",
    "KURDISH": "KU",
    "KYRGYZ": "KY",
    "LAO": "LO",
    "LATIN": "LA",
    "LATVIAN": "LV",
    "LITHUANIAN": "LT",
    "LUXEMBOURGISH": "LB",
    "MACEDONIAN": "MK",
    "MALAGASY": "MG",
    "MALAY": "MS",
    "MALAYALAM": "ML",
    "MALTESE": "MT",
    "MAORI": "MI",
    "MARATHI": "MR",
    "MONGOLIAN": "MN",
    "MYANMAR": "MY",
    "BURMESE": "MY",
    "NEPALI": "NE",
    "NORWEGIAN": "NO",
    "NYANJA": "NY",
    "CHICHEWA": "NY",
    "PASHTO": "PS",
    "PERSIAN": "FA",
    "POLISH": "PL",
    "PORTUGUESE": "PT",
    "BRAZILIAN": "PT",
    "PUNJABI": "PA",
    "ROMANIAN": "RO",
    "RUSSIAN": "RU",
    "SAMOAN": "SM",
    "SCOTTISH": "GD",
    "SCOTS GAELIC": "GD",
    "SERBIAN": "SR",
    "SESOTHO": "ST",
    "SHONA": "SN",
    "SINDHI": "SD",
    "SINHALA": "SI",
    "SINHALESE": "SI",
    "SLOVAK": "SK",
    "SLOVENIAN": "SL",
    "SOMALI": "SO",
    "SPANISH": "ES",
    "SUNDANESE": "SU",
    "SWAHILI": "SW",
    "SWEDISH": "SV",
    "TAGALOG": "TL",
    "FILIPINO": "TL",
    "TAJIK": "TG",
    "TAMIL": "TA",
    "TELUGU": "TE",
    "THAI": "TH",
    "TURKISH": "TR",
    "UKRAINIAN": "UK",
    "URDU": "UR",
    "UZBEK": "UZ",
    "VIETNAMESE": "VI",
    "WELSH": "CY",
    "XHOSA": "XH",
    "YIDDISH": "YI",
    "YORUBA": "YO",
    "ZULU": "ZU"
}


class TextBlobTranslator(BaseTranslator):

    def languages(self):
        return ", ".join(languages.keys())

    def supports_language(self, language):
        return bool(language in languages)

    def language_code(self, code):
        if code in languages:
            return languages[code]
        return "UNKNOWN"

    def _get_textblob(self, text):
        return TextBlob(text)

    def detect(self, text, default='EN'):
        blob = self._get_textblob(text)
        try:
            return blob.detect_language().upper()

        except TranslatorError as te:
            YLogger.exception(None, "Unable to determine langauge for [%s]", te, text)

        return default

    def _do_translate(self, text, from_lang, to_lang):
        blob = TextBlob(text)

        if from_lang is not None:
            return str(blob.translate(from_lang=from_lang, to=to_lang))

        else:
            return str(blob.translate(to=to_lang))

    def translate(self, text, from_lang=None, to_lang="EN"):

        YLogger.debug(None, "Translating [%s] from [%s] to [%s], are they the same?", text, from_lang, to_lang)

        try:
            translated = self._do_translate(text, from_lang, to_lang)
            YLogger.debug(None, "Translated [%s] to [%s]", text, translated)
            return translated

        except NotTranslated as nte:
            YLogger.exception(None, "Unable to translate text from [%s] to [%s], are they the same?",
                              nte, from_lang, to_lang)

        except URLError as urle:
            YLogger.exception(None, "No connection to Google Translate", urle)

        return text



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import os
import os.path
import shutil
from programy.utils.logging.ylogger import YLogger
from programy.storage.entities.store import Store
from programy.utils.console.console import outputLog


class FileStore(Store):
    FILE = "file"

    CATEGORIES_STORAGE = 'categories_storage'
    ERRORS_STORAGE = 'errors_storage'
    DUPLICATES_STORAGE = 'duplicates_storage'
    LEARNF_STORAGE = 'learnf_storage'

    CONVERSATION_STORAGE = 'conversation_storage'

    SETS_STORAGE = 'sets_storage'
    MAPS_STORAGE = 'maps_storage'
    RDF_STORAGE = 'rdf_storage'

    DENORMAL_STORAGE = 'denormal_storage'
    NORMAL_STORAGE = 'normal_storage'
    GENDER_STORAGE = 'gender_storage'
    PERSON_STORAGE = 'person_storage'
    PERSON2_STORAGE = 'person2_storage'
    REGEX_STORAGE = 'regex_storage'

    PROPERTIES_STORAGE = 'properties_storage'
    VARIABLES_STORAGE = 'variables_storage'
    DEFAULTS_STORAGE = 'defaults_storage'

    TWITTER_STORAGE = 'twitter_storage'

    SPELLING_STORAGE = 'spelling_storage'

    LICENSE_STORAGE = 'license_storage'

    PATTERN_NODES_STORAGE = 'pattern_nodes_storage'
    TEMPLATE_NODES_STORAGE = 'template_nodes_storage'

    BINARIES_STORAGE = 'binaries_storage'
    BRAINTREE_STORAGE = 'braintree_storage'

    PREPROCESSORS_STORAGE = 'preprocessors_storage'
    POSTPROCESSORS_STORAGE = 'postprocessors_storage'
    POSTQUESTIONPROCESSORS_STORAGE = 'postquestionprocessors_storage'

    USERGROUPS_STORAGE = 'usergroups_storage'

    TRIGGERS_STORAGE = 'triggers_storage'

    OOBS_STORAGE = 'oobs_storage'

    SERVICES_STORAGE = 'services_storage'

    def __init__(self, storage_engine):
        self._storage_engine = storage_engine

    def empty(self):
        pass    # pragma: no cover

    def empty_named(self, name):
        pass    # pragma: no cover

    @property
    def storage_engine(self):
        return self._storage_engine

    def _get_storage_path(self):
        raise NotImplementedError("Implement _get_storage_path to return storage specific "
                                  "folder from config")  # pragma: no cover

    def _drop_folder(self, folder):
        if os.path.exists(folder):
            shutil.rmtree(folder)

    def drop(self):
        try:
            self._drop_folder(self._get_storage_path())
        except Exception as e:
            YLogger.exception_nostack(self, "Error dropping storage", e)

    @staticmethod
    def _get_dir_from_path(file_path):
        splits = file_path.split(os.sep)
        paths = splits[:len(splits) - 1]
        path = "/".join(paths)
        return path

    @staticmethod
    def _ensure_dir_exists(path):
        if os.path.exists(path) is False:
            YLogger.debug(None, "Directory does not exist, creating %s", path)
            os.makedirs(path)

    @staticmethod
    def _file_exists(path):
        return os.path.exists(path)

    def commit(self, commit=True):
        del commit
        pass    # pragma: no cover

    def load_all(self, collector):
        col_storage = self.get_storage()

        if col_storage.has_multiple_dirs():
            subdir = col_storage.subdirs
            col_ext = col_storage.extension
            for col_dir in col_storage.dirs:
                if subdir is False:
                    paths = os.listdir(col_dir)
                    for filename in paths:
                        if col_ext is None or filename.endswith(col_ext):
                            YLogger.debug(self, "Loading file contents from [%s]", filename)
                            self._load_file_contents(collector, os.path.join(col_dir, filename))
                else:
                    for dirpath, _, filenames in os.walk(col_dir):
                        for filename in [f for f in filenames if f.endswith(col_ext)]:
                            YLogger.debug(self, "Loading file contents from [%s]", filename)
                            self._load_file_contents(collector, os.path.join(dirpath, filename))

        else:
            self.load(collector)

    def load(self, collector, name=None):
        del name
        col_storage = self.get_storage()
        collector.empty()
        self._load_file_contents(collector, col_storage.file)

    def _load_file_contents(self, collection, filename):
        pass    # pragma: no cover

    def get_storage(self):
        raise NotImplementedError("get_storage must be implemented in child class")  # pragma: no cover



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from programy.utils.logging.ylogger import YLogger
from programy.storage.stores.nosql.mongo.store.mongostore import MongoStore
from programy.storage.entities.lookups import LookupsStore
from programy.storage.stores.nosql.mongo.dao.lookup import Lookup
from programy.mappings.base import DoubleStringPatternSplitCollection


class MongoLookupStore(MongoStore, LookupsStore):

    def __init__(self, storage_engine):
        MongoStore.__init__(self, storage_engine)
        LookupsStore.__init__(self)

    def collection_name(self):
        raise NotImplementedError()  # pragma: no cover

    def add_to_lookup(self, key, value, overwrite_existing=False):
        collection = self.collection()
        lookup = collection.find_one({'key': key})
        if lookup is not None:
            if overwrite_existing is True:
                YLogger.info(self, "Updating lookup in Mongo [%s] [%s]", key, value)
                lookup['value'] = value
                result = collection.replace_one({'key': key}, lookup)
                return bool(result.modified_count > 0)

            else:
                YLogger.error(self, "Existing value in Mongo lookup [%s] = [%s]", key, value)

        else:
            YLogger.debug(self, "Adding lookup to Mongo [%s] = [%s]", key, value)
            lookup = Lookup(key, value)
            return self.add_document(lookup)

        return False

    def remove_lookup(self):
        YLogger.debug(self, "Removing lookup from Mongo [%s]", self.collection_name())
        collection = self.collection()
        collection.delete_many({})

    def remove_lookup_key(self, key):
        YLogger.debug(self, "Removing lookup key [%s] from [%s] in Mongo", key, self.collection_name())
        collection = self.collection()
        collection.delete_one({'key': key})

    def get_lookup(self):
        collection = self.collection()
        lookups = collection.find()
        data = {}
        for lookup in lookups:
            data[lookup['key']] = lookup['value']

        return data

    def load_all(self, collector):
        self.load(collector)

    def load(self, collector, name=None):
        YLogger.debug(self, "Loading lookup from Mongo [%s]", self.collection_name())
        collection = self.collection()
        lookups = collection.find()
        for lookup in lookups:
            key, value = DoubleStringPatternSplitCollection.process_key_value(lookup['key'], lookup['value'])
            collector.add_to_lookup(key, value)

        return True

    def process_line(self, name, fields, verbose=False):
        if fields:
            key = fields[0].upper()
            value = fields[1]
            return self.add_to_lookup(key, value)
        return False


class MongoDenormalStore(MongoLookupStore):

    def __init__(self, storage_engine):
        MongoLookupStore.__init__(self, storage_engine)

    def collection_name(self):
        return "denormals"


class MongoNormalStore(MongoLookupStore):

    def __init__(self, storage_engine):
        MongoLookupStore.__init__(self, storage_engine)

    def collection_name(self):
        return "normals"


class MongoGenderStore(MongoLookupStore):

    def __init__(self, storage_engine):
        MongoLookupStore.__init__(self, storage_engine)

    def collection_name(self):
        return "genders"


class MongoPersonStore(MongoLookupStore):

    def __init__(self, storage_engine):
        MongoLookupStore.__init__(self, storage_engine)

    def collection_name(self):
        return "persons"


class MongoPerson2Store(MongoLookupStore):

    def __init__(self, storage_engine):
        MongoLookupStore.__init__(self, storage_engine)

    def collection_name(self):
        return "person2s"



import xml.etree.ElementTree as ET
from os import walk, sep

file_converters = {
    "DefaultPredicates":    { "seperator": ":", "filename": "defaults.txt"},
    "GenderSubstitutions":  { "seperator": ":", "filename": "gender.txt"},
    "PersonSubstitutions":  { "seperator": ":", "filename": "person.txt"},
    "Person2Substitutions": { "seperator": ":", "filename": "person2.txt"},
    "Splitters":            {                   "filename": "converters.txt"},
    "Substitutions":        { "seperator": ":", "filename": "denormals.txt"},
    "Settings":             { "seperator": ":", "filename": "settings.txt", "ignores": [
                                                                                        "aimldirectory",
                                                                                        "configdirectory",
                                                                                        "logdirectory",
                                                                                        "convertersfile",
                                                                                        "person2substitutionsfile",
                                                                                        "personsubstitutionsfile",
                                                                                        "gendersubstitutionsfile",
                                                                                        "defaultpredicates",
                                                                                        "substitutionsfile",
                                                                                        "maxlogbuffersize",
                                                                                        "notacceptinguserinputmessage",
                                                                                        "stripperregex",
                                                                                        "islogging",
                                                                                        "willcallhome",
                                                                                        "timeout",
                                                                                        "timeoutmessage",]}
    }


class PandoraBotsFileConverter(object):

    @staticmethod
    def all_files(path):
        files = []
        for (_, _, names) in walk(path):
            for name in names:
                files.append(path + sep + name)
            break
        return files

    @staticmethod
    def get_filename(file):
        splits1 = file.split(sep)
        splits2 = splits1[-1].split(".")
        return splits2[0]

    @staticmethod
    def convert_file(from_file, to_file, converter):
        print(from_file)
        with open(to_file, "w+") as open_file:
            tree = ET.parse(from_file)
            xml = tree.getroot()
            for item in xml:
                name = None
                if 'name' in item.attrib:
                    name = item.attrib['name'].strip()
                value = item.attrib['value'].strip()

                seperator = ":"
                if 'seperator' in converter:
                    seperator = converter['seperator']

                ignores = []
                if 'ignores' in converter:
                    ignores = converter['ignores']

                if name:
                    if name not in ignores:
                        open_file.write("%s%s%s\n" % (name, seperator, value))
                else:
                    open_file.write("%s\n" % value)

    def convert(self, from_files, to_files):

        files = self.all_files(from_files)
        for from_file in files:
            filename = self.get_filename(from_file)
            if filename in file_converters:
                converter = file_converters[filename]
                to_file = to_files + sep + converter["filename"]
                self.convert_file(from_file, to_file, converter)


if __name__ == '__main__':

    from_files = "./pbfiles"
    to_files = "./yfiles"
    converter = ":"

    converter = PandoraBotsFileConverter()
    converter.convert(from_files, to_files)


"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from programy.utils.logging.ylogger import YLogger
from programy.utils.console.console import outputLog
from programy.clients.events.client import EventBotClient
from programy.clients.events.console.config import ConsoleConfiguration


class ConsoleBotClient(EventBotClient):

    def __init__(self, argument_parser=None):
        self.running = False
        EventBotClient.__init__(self, "Console", argument_parser)

    def _render_callback(self):
        return True

    def get_client_configuration(self):
        return ConsoleConfiguration()

    def add_client_arguments(self, parser=None):
        return

    def parse_args(self, arguments, parsed_args):
        return

    def get_question(self, client_context, input_func=None):
        del client_context
        ask = "%s " % self._configuration.client_configuration.prompt
        if input_func:
            return input_func(ask)

        return input(ask)       #pylint: disable=input-builtin

    def display_startup_messages(self, client_context):
        self.process_response(client_context, client_context.bot.get_version_string(client_context))
        initial_question = client_context.bot.get_initial_question(client_context)
        self._renderer.render(client_context, initial_question)

    def process_question(self, client_context, question):
        self._questions += 1
        return client_context.bot.ask_question(client_context, question, responselogger=self)

    def render_response(self, client_context, response):
        # Calls the renderer which handles RCS context, and then calls back to the client to show response
        self._renderer.render(client_context, response)

    def process_response(self, client_context, response):
        del client_context
        outputLog(self, response)

    def process_question_answer(self, client_context):
        question = self.get_question(client_context)
        response = self.process_question(client_context, question)
        self.render_response(client_context, response)

    def wait_and_answer(self):
        running = True
        try:
            client_context = self.create_client_context(self._configuration.client_configuration.default_userid)
            self.process_question_answer(client_context)

        except KeyboardInterrupt:
            running = False
            client_context = self.create_client_context(self._configuration.client_configuration.default_userid)
            self._renderer.render(client_context, client_context.bot.get_exit_response(client_context))

        except Exception as excep:
            YLogger.exception(self, "Oops something bad happened !", excep)

        return running

    def prior_to_run_loop(self):
        client_context = self.create_client_context(self._configuration.client_configuration.default_userid)
        self.display_startup_messages(client_context)


if __name__ == '__main__':

    outputLog(None, "Initiating Console Client...")

    def run():
        console_app = ConsoleBotClient()
        console_app.run()

    run()



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

import re
import os


class TextUtils:
    DEFAULT_TAB_SPACE = "  "

    RE_STRIP_ALL_WHITESPACE = re.compile(r'[\s+]')
    RE_STRIP_WHITESPACE = re.compile(r'[\n\t\r+]')
    RE_TERMIANTORS = re.compile(r'[:;,.?!]')
    RE_STRIP_NONE_TERMINATING = re.compile(r'[\'"\(\)\-"]')
    RE_STRIP_ALL_PUNCTUATION1 = re.compile(r'[:\'";,.?!\(\)\-"]')
    RE_STRIP_ALL_PUNCTUATION2 = re.compile(r'\s+')
    RE_STRIP_HTML = re.compile(r'<.*?>')
    RE_PATTERN_OF_TAG_AND_NAMESPACE_FROM_TEXT = re.compile(r'^{.+}.+$')
    RE_MATCH_OF_TAG_AND_NAMESPACE_FROM_TEXT = re.compile(r'^({.+})(.+)$')

    HTML_ESCAPE_TABLE = {
        "&": "&amp;",
        '"': "&quot;",
        "'": "&apos;",
        ">": "&gt;",
        "<": "&lt;",
    }

    @staticmethod
    def get_tabs(depth: int, tabs=DEFAULT_TAB_SPACE):
        return tabs * depth

    @staticmethod
    def strip_whitespace(string):
        first_pass = TextUtils.RE_STRIP_WHITESPACE.sub(' ', string)
        return ' '.join(first_pass.split())

    @staticmethod
    def strip_all_whitespace(string):
        first_pass = TextUtils.RE_STRIP_ALL_WHITESPACE.sub('', string)
        return ' '.join(first_pass.split())

    @staticmethod
    def strip_none_terminating_punctuation(string):
        first_pass = TextUtils.RE_STRIP_NONE_TERMINATING.sub(' ', string)
        second_pass = TextUtils.RE_STRIP_NONE_TERMINATING.sub(' ', first_pass)
        return second_pass.strip()

    @staticmethod
    def strip_all_punctuation(string):
        first_pass = TextUtils.RE_STRIP_ALL_PUNCTUATION1.sub(' ', string)
        second_pass = TextUtils.RE_STRIP_ALL_PUNCTUATION2.sub(' ', first_pass)
        return second_pass.strip()

    @staticmethod
    def urlify(string):
        return string.replace(" ", "%20")

    @staticmethod
    def strip_html(data, replace_with=''):
        return TextUtils.RE_STRIP_HTML.sub(replace_with, data)

    @staticmethod
    def replace_path_seperator(path, old="/", new=os.sep):
        if old in path:
            return path.replace(old, new)
        return path

    @staticmethod
    def _is_not_tag(text):
        # If there is a namespace, then it looks something like
        # {http://alicebot.org/2001/AIML}aiml
        return bool(TextUtils.RE_PATTERN_OF_TAG_AND_NAMESPACE_FROM_TEXT.match(text) is None)

    @staticmethod
    def _extract_namespace_and_tag(text):
        # Otherwise, extract namespace and tag name
        groupings = TextUtils.RE_MATCH_OF_TAG_AND_NAMESPACE_FROM_TEXT.match(text)
        if groupings is not None:
            namespace = groupings.group(1).strip()
            tag_name = groupings.group(2).strip()
            return tag_name, namespace

        return None, None

    @staticmethod
    def tag_and_namespace_from_text(text):
        if TextUtils._is_not_tag(text):
            return text, None

        return TextUtils._extract_namespace_and_tag(text)

    @staticmethod
    def tag_from_text(text):
        tag, _ = TextUtils.tag_and_namespace_from_text(text)
        return tag

    @staticmethod
    def html_escape(text):
        return "".join(TextUtils.HTML_ESCAPE_TABLE.get(c, c) for c in text)



try:
    from Tkinter import Tk
    from ttk import Treeview, Scrollbar, Frame
    from Tkconstants import HORIZONTAL, VERTICAL, N, S, E, W, END, BOTH, YES
except ImportError:
    from tkinter import Tk
    from tkinter.ttk import Treeview, Scrollbar, Frame
    from tkinter.constants import HORIZONTAL, VERTICAL, N, S, E, W, END, BOTH, YES

import xml.etree.ElementTree as ET
import sys
from operator import attrgetter

def autoscroll(sbar, first, last):
    """Hide and show scrollbar as needed."""
    first, last = float(first), float(last)
    if first <= 0 and last >= 1:
        sbar.grid_remove()
    else:
        sbar.grid()
    sbar.set(first, last)


class XML_Viewer(Frame):
    def __init__(self, master, xml=None, heading_text=None, heading_anchor=None, padding=None, cursor=None,
                 takefocus=None, style=None):
        Frame.__init__(self, master, class_="XML_Viewer")

        self._vsb = Scrollbar(self, orient=VERTICAL)
        self._hsb = Scrollbar(self, orient=HORIZONTAL)

        kwargs = {}
        kwargs["yscrollcommand"] = lambda f, l: autoscroll(self._vsb, f, l)
        kwargs["xscrollcommand"] = lambda f, l: autoscroll(self._hsb, f, l)

        if style is not None:
            kwargs["style"] = style

        if padding is not None:
            kwargs["padding"] = padding

        if cursor is not None:
            kwargs["cursor"] = cursor

        if takefocus is not None:
            kwargs["takefocus"] = takefocus

        self._treeview = Treeview(self, **kwargs)

        if heading_text is not None:
            if heading_anchor is not None:
                self._treeview.heading("#0", text=heading_text, anchor=heading_anchor)
            else:
                self._treeview.heading("#0", text=heading_text)

        self._treeview.bind("<<TreeviewOpen>>", self._on_open)
        self._treeview.bind("<<TreeviewClose>>", self._on_close)

        # Without this line, horizontal scrolling doesn't work properly.
        self._treeview.column("#0", stretch=False)
        self._treeview.column("#0", width=500)

        self._vsb['command'] = self._treeview.yview
        self._hsb['command'] = self._treeview.xview

        self._treeview.grid(column=0, row=0, sticky=N + S + W + E)
        self._vsb.grid(column=1, row=0, sticky=N + S)
        self._hsb.grid(column=0, row=1, sticky=E + W)

        self.grid_columnconfigure(0, weight=1)
        self.grid_rowconfigure(0, weight=1)

        self._element_tree = None
        self._item_ID_to_element = {}

        if xml is not None:
            self.parse_xml(xml)

    def _on_open(self, event):
        del event
        item_ID = self._treeview.focus()
        if item_ID not in self._item_ID_to_element: return

        node = self._item_ID_to_element[item_ID]

        self._treeview.item(item_ID, text=self._repr_of_openning_tag(node))

    def _on_close(self, event):
        del event
        item_ID = self._treeview.focus()
        if item_ID not in self._item_ID_to_element: return

        node = self._item_ID_to_element[item_ID]

        text = self._repr_of_openning_tag(node) + self._repr_of_closing_tag(node)
        self._treeview.item(item_ID, text=text)

    def parse_xml(self, xml):
        self._element_tree = ET.ElementTree(ET.fromstring(xml))

        self.clear()
        self._walk_xml(self._element_tree.getroot())

    @property
    def element_tree(self):
        return self._element_tree

    @element_tree.setter
    def element_tree(self, element_tree):
        self._element_tree = element_tree

        self.clear()
        self._walk_xml(element_tree.getroot())

    def clear(self):
        self._item_ID_to_element = {}
        self._treeview.delete(*self._treeview.get_children())

    def _repr_of_openning_tag(self, node):
        text = "<" + node.tag

        attrs = node.attrib

        # list function is here necessary to provide support to Python 3
        a_names = list(attrs.keys())
        a_names.sort()

        for a_name in a_names:
            text += ' %s="' % a_name
            text += attrs[a_name]
            text += '"'

        text += ">"
        return text

    def _repr_of_closing_tag(self, node):
        return "</%s>" % node.tag

    def _walk_xml(self, node, depth=0, parent=""):
        text = self._repr_of_openning_tag(node) + self._repr_of_closing_tag(node)

        item = self._treeview.insert(parent, END, text=text)
        self._item_ID_to_element[item] = node

        if node.text:
            text = node.text.strip()
            if text != "":
                for line in text.splitlines():
                    self._treeview.insert(item, END, text=line)

        sorted_nodes = sorted(list(node), key=attrgetter('tag'))
        for child_node in sorted_nodes:
            self._walk_xml(child_node, depth + 1, parent=item)

        if node.tail:
            tail = node.tail.strip()
            if tail != "":
                for line in tail.splitlines():
                    self._treeview.insert(parent, END, text=line)


if __name__ == "__main__":

    if len(sys.argv) != 2 or sys.argv[1] is None or sys.argv[1] == "":
        print("No xml file name specified!")
        exit(0)

    filename = sys.argv[1]

    root = Tk()
    with open(filename, "r") as xml_file:
        xml = xml_file.read()
    XML_Viewer(root, xml, heading_text="Braintree").pack(fill=BOTH, expand=YES)
    root.mainloop()



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from programy.utils.logging.ylogger import YLogger
from programy.storage.stores.nosql.mongo.store.mongostore import MongoStore
from programy.storage.entities.maps import MapsReadWriteStore
from programy.storage.stores.nosql.mongo.dao.map import Map


class MongoMapsStore(MongoStore, MapsReadWriteStore):
    MAPS = 'maps'
    NAME = 'name'
    KEYVALUES = 'key_values'

    def __init__(self, storage_engine):
        MongoStore.__init__(self, storage_engine)
        MapsReadWriteStore.__init__(self)

    def collection_name(self):
        return MongoMapsStore.MAPS

    def empty_named(self, name):
        YLogger.info(self, "Empting map in Mongo [%s]", name)
        collection = self.collection()
        collection.remove({MongoMapsStore.NAME: name})

    def add_to_map(self, name, key, value, overwrite_existing=False):
        collection = self.collection()
        amap = collection.find_one({MongoMapsStore.NAME: name})
        if amap is not None:
            if key in amap[MongoMapsStore.KEYVALUES]:
                if overwrite_existing is True:
                    YLogger.info(self, "Updating map [%s] in Mongo [%s]=[%s]", name, key, value)
                    amap[MongoMapsStore.KEYVALUES][key] = value
                    result = collection.replace_one({MongoMapsStore.NAME: name}, amap)
                    return bool(result.modified_count > 0)
                else:
                    YLogger.error(self, "Existing value in map [%s] [%s] = [%s] in Mongo", name, key, value)
            else:
                amap[MongoMapsStore.KEYVALUES][key] = value
                result = collection.replace_one({MongoMapsStore.NAME: name}, amap)
                return bool(result.modified_count > 0)

        else:
            amap = Map(name, {key: value})
            return self.add_document(amap)

        return False

    def remove_from_map(self, name, key):
        YLogger.info(self, "Removing key [%s] from map [%s] in Mongo", name, key)
        collection = self.collection()
        amap = collection.find_one({MongoMapsStore.NAME: name})
        if amap is not None:
            amap[MongoMapsStore.KEYVALUES].pop(key)
            if amap[MongoMapsStore.KEYVALUES]:
                result = collection.replace_one({MongoMapsStore.NAME: name}, amap)
                return bool(result.modified_count > 0)

            else:
                result = collection.delete_one({MongoMapsStore.NAME: name})
                return bool(result.deleted_count > 0)

        return False

    def load_all(self, collector):
        YLogger.info(self, "Loading all maps from Mongo")
        collection = self.collection()
        collector.empty()
        maps = collection.find({})
        for amap in maps:
            self.load(collector, amap[MongoMapsStore.NAME])

    def load(self, collector, name=None):
        YLogger.info(self, "Loading map [%s] from Mongo", name)
        collection = self.collection()
        amap = collection.find_one({MongoMapsStore.NAME: name})
        if amap is not None:
            collector.remove(name)
            collector.add_map(name, amap[MongoMapsStore.KEYVALUES], MongoStore.MONGO)
            return True

        return False



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from pymongo import MongoClient
from programy.storage.engine import StorageEngine
from programy.storage.stores.nosql.mongo.store.categories import MongoCategoryStore
from programy.storage.stores.nosql.mongo.store.learnf import MongoLearnfStore
from programy.storage.stores.nosql.mongo.store.errors import MongoErrorsStore
from programy.storage.stores.nosql.mongo.store.duplicates import MongoDuplicatesStore
from programy.storage.stores.nosql.mongo.store.conversations import MongoConversationStore
from programy.storage.stores.nosql.mongo.store.lookups import MongoDenormalStore
from programy.storage.stores.nosql.mongo.store.lookups import MongoNormalStore
from programy.storage.stores.nosql.mongo.store.lookups import MongoGenderStore
from programy.storage.stores.nosql.mongo.store.lookups import MongoPersonStore
from programy.storage.stores.nosql.mongo.store.lookups import MongoPerson2Store
from programy.storage.stores.nosql.mongo.store.properties import MongoPropertyStore
from programy.storage.stores.nosql.mongo.store.properties import MongoDefaultVariablesStore
from programy.storage.stores.nosql.mongo.store.properties import MongoRegexesStore
from programy.storage.stores.nosql.mongo.store.twitter import MongoTwitterStore
from programy.storage.stores.nosql.mongo.store.sets import MongoSetsStore
from programy.storage.stores.nosql.mongo.store.maps import MongoMapsStore
from programy.storage.stores.nosql.mongo.store.rdfs import MongoRDFsStore
from programy.storage.stores.nosql.mongo.store.spelling import MongoSpellingStore
from programy.storage.stores.nosql.mongo.store.licensekeys import MongoLicenseKeysStore
from programy.storage.stores.nosql.mongo.store.nodes import MongoPatternNodeStore
from programy.storage.stores.nosql.mongo.store.nodes import MongoTemplateNodeStore
from programy.storage.stores.nosql.mongo.store.processors import MongoPreProcessorStore
from programy.storage.stores.nosql.mongo.store.processors import MongoPostProcessorStore
from programy.storage.stores.nosql.mongo.store.processors import MongoPostQuestionProcessorStore
from programy.storage.stores.nosql.mongo.store.usergroups import MongoUserGroupsStore
from programy.storage.stores.nosql.mongo.store.users import MongoUserStore
from programy.storage.stores.nosql.mongo.store.linkedaccounts import MongoLinkedAccountStore
from programy.storage.stores.nosql.mongo.store.links import MongoLinkStore
from programy.storage.stores.nosql.mongo.store.triggers import MongoTriggerStore
from programy.storage.stores.nosql.mongo.store.oobs import MongoOOBStore


class MongoStorageEngine(StorageEngine):

    def __init__(self, configuration):
        StorageEngine.__init__(self, configuration)
        self._client = None
        self._database = None

    @property
    def database(self):
        return self._database

    def initialise(self):
        self._client = MongoClient(self.configuration.url)

        if self.configuration.drop_all_first is True:
            self._client.drop_database(self.configuration.database)

        self._database = self._client.get_database(self.configuration.database)

        return True

    def category_store(self):
        return MongoCategoryStore(self)

    def learnf_store(self):
        return MongoLearnfStore(self)

    def errors_store(self):
        return MongoErrorsStore(self)

    def duplicates_store(self):
        return MongoDuplicatesStore(self)

    def conversation_store(self):
        return MongoConversationStore(self)

    def sets_store(self):
        return MongoSetsStore(self)

    def maps_store(self):
        return MongoMapsStore(self)

    def rdf_store(self):
        return MongoRDFsStore(self)

    def denormal_store(self):
        return MongoDenormalStore(self)

    def normal_store(self):
        return MongoNormalStore(self)

    def gender_store(self):
        return MongoGenderStore(self)

    def person_store(self):
        return MongoPersonStore(self)

    def person2_store(self):
        return MongoPerson2Store(self)

    def property_store(self):
        return MongoPropertyStore(self)

    def defaults_store(self):
        return MongoDefaultVariablesStore(self)

    def regex_store(self):
        return MongoRegexesStore(self)

    def twitter_store(self):
        return MongoTwitterStore(self)

    def spelling_store(self):
        return MongoSpellingStore(self)

    def license_store(self):
        return MongoLicenseKeysStore(self)

    def pattern_nodes_store(self):
        return MongoPatternNodeStore(self)

    def template_nodes_store(self):
        return MongoTemplateNodeStore(self)

    def preprocessors_store(self):
        return MongoPreProcessorStore(self)

    def postprocessors_store(self):
        return MongoPostProcessorStore(self)

    def postquestionprocessors_store(self):
        return MongoPostQuestionProcessorStore(self)

    def usergroups_store(self):
        return MongoUserGroupsStore(self)

    def user_store(self):
        return MongoUserStore(self)

    def linked_account_store(self):
        return MongoLinkedAccountStore(self)

    def link_store(self):
        return MongoLinkStore(self)

    def triggers_store(self):
        return MongoTriggerStore(self)

    def oobs_store(self):
        return MongoOOBStore(self)



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from programy.utils.logging.ylogger import YLogger
from programy.storage.stores.file.store.filestore import FileStore
from programy.storage.entities.braintree import BraintreeStore


class FileBraintreeStore(FileStore, BraintreeStore):

    def __init__(self, storage_engine):
        FileStore.__init__(self, storage_engine)
        BraintreeStore.__init__(self)

    def _get_storage_path(self):
        return self.storage_engine.configuration.braintree_storage.file

    def get_storage(self):
        return self.storage_engine.configuration.braintree_storage

    def _save(self, client_context, pattern_graph):

        braintree_fullpath = self.storage_engine.configuration.braintree_storage.file
        YLogger.info(self, "Saving braintree to %s", braintree_fullpath)

        braintree_dirpath = self._get_dir_from_path(braintree_fullpath)
        self._ensure_dir_exists(braintree_dirpath)

        fileformat = self.storage_engine.configuration.braintree_storage.format
        encoding = self.storage_engine.configuration.braintree_storage.encoding

        if fileformat == FileStore.TEXT_FORMAT:
            with open(braintree_fullpath, "w+", encoding=encoding) as dump_file:
                pattern_graph.dump(output_func=dump_file.write, eol="\n")

        elif fileformat == FileStore.XML_FORMAT:
            braintree = '<?xml version="1.0" encoding="%s"?>\n' % encoding
            braintree += '<aiml>\n'
            braintree += pattern_graph.root.to_xml(client_context)
            braintree += '</aiml>\n'
            with open(braintree_fullpath, "w+", encoding=encoding) as dump_file:
                dump_file.write(braintree)

        else:
            YLogger.error(client_context, "Unknown braintree content type [%s]", fileformat)

    def save_braintree(self, client_context, pattern_graph):

        try:
            self._save(client_context, pattern_graph)

        except Exception as exc:
            YLogger.exception_nostack(client_context, "Failed to save Braintree", exc)



from setuptools import setup, find_packages
from os import path

here = path.abspath(path.dirname(__file__))

version = None
with open(path.join(here, 'version.txt'), encoding='utf-8') as f:
    version = f.read()
version = version.strip()
print("Building PyPi Package for version %s"%version)

if version is None:
    print("No version.txt found")

long_description = ""
with open(path.join(here, path.join(here, 'README.md')), encoding='utf-8') as f:
    long_description = f.read()

download_url = "https://github.com/keiffster/program-y/%s.tar.gz"%version

setup(
  name = 'programy',
  packages=find_packages(),
  package_data={'': ['*.conf', '*.aiml']},
  include_package_data=True,
  version = version,
  description = 'AIML Framework and Platform',
  long_description=long_description,
  long_description_content_type='text/markdown',
  author = 'Keith Sterling',
  author_email = 'keiffster@gmail.com',
  url = 'https://github.com/keiffster/program-y.git',
  download_url = download_url,
  keywords = ['aiml', 'chatbot', 'virtual assistant', 'ai'],
  classifiers = [
      # How mature is this project? Common values are
      #   3 - Alpha
      #   4 - Beta
      #   5 - Production/Stable
      'Development Status :: 5 - Production/Stable',

      # Indicate who your project is intended for
      'Intended Audience :: Developers',

      # Pick your license as you wish (should match "license" above)
      'License :: OSI Approved :: MIT License',

      # Specify the Python versions you support here. In particular, ensure
      # that you indicate whether you support Python 2, Python 3 or both.
      'Programming Language :: Python :: 3',
      'Programming Language :: Python :: 3.2',
      'Programming Language :: Python :: 3.3',
      'Programming Language :: Python :: 3.4',
      'Programming Language :: Python :: 3.5',
      'Programming Language :: Python :: 3.6',
      'Programming Language :: Python :: 3.7',
  ],
  install_requires=[
      'python-dateutil==2.8.1',
      'wget==3.2',
      'beautifulsoup4==4.8',
      'lxml==4.4.2',
      'PyYAML==5.3b1',
      'requests==2.22.0',
      'Flask==1.1.1',
      'Flask-SocketIO==4.2.1',
      'tweepy==3.8.0',
      'sleekxmpp==1.3.3',
      'python-telegram-bot==12.0.0',
      'pymessenger==0.0.7.0',
      'twilio==6.35.1',
      'slackclient==1.3.1',
      'viberbot==1.0.11',
      'line-bot-sdk==1.15.0',
      'kik==1.5.0',
      'discord.py==1.2.3',
      'botbuilder-core==4.5.0b4',
      'botbuilder-schema>=4.5.0b4',
      'wikipedia==1.4.0',
      'MetOffer==1.3.2',
      'APScheduler==3.6.3',
      'emoji==0.5.4',
      'autocorrect==0.4.4',
      'textblob==0.15.3',
      'redis==3.3.11',
      'pymongo==3.10.0',
      'SQLAlchemy==1.3.12',
      'PyMySQL==0.9.3'
]
)


"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import argparse
import yaml

from programy.clients.events.console.config import ConsoleConfiguration
from programy.clients.events.tcpsocket.config import SocketConfiguration
from programy.clients.polling.slack.config import SlackConfiguration
from programy.clients.polling.telegram.config import TelegramConfiguration
from programy.clients.polling.twitter.config import TwitterConfiguration
from programy.clients.polling.xmpp.config import XmppConfiguration
from programy.clients.restful.config import RestConfiguration
from programy.clients.restful.flask.facebook.config import FacebookConfiguration
from programy.clients.restful.flask.kik.config import KikConfiguration
from programy.clients.restful.flask.line.config import LineConfiguration
from programy.clients.restful.flask.twilio.config import TwilioConfiguration
from programy.clients.restful.flask.viber.config import ViberConfiguration
from programy.clients.restful.sanic.config import SanicRestConfiguration
from programy.utils.console.console import outputLog


class ConfigurationWriter:

    def __init__(self):
        pass    # pragma: no cover

    def add_to_config(self, config_data, configuration, defaults=True):
        config_data[configuration.id] = {}
        configuration.to_yaml(config_data[configuration.id], defaults)

    def execute(self, args):

        config_data = {}

        if args is None:
            raise Exception("Args empty")

        if args.clients is None:
            raise Exception("No clients defined")

        if 'all' in args.clients or 'console' in args.clients:
            self.add_to_config(config_data, ConsoleConfiguration(), args.defaults)

        if 'all' in args.clients or 'socket' in args.clients:
            self.add_to_config(config_data, SocketConfiguration(), args.defaults)

        if 'all' in args.clients or 'slack' in args.clients:
            self.add_to_config(config_data, SlackConfiguration(), args.defaults)

        if 'all' in args.clients or 'telegram' in args.clients:
            self.add_to_config(config_data, TelegramConfiguration(), args.defaults)

        if 'all' in args.clients or 'twitter' in args.clients:
            self.add_to_config(config_data, TwitterConfiguration(), args.defaults)

        if 'all' in args.clients or 'xmpp' in args.clients:
            self.add_to_config(config_data, XmppConfiguration(), args.defaults)

        if 'all' in args.clients or 'rest' in args.clients:
            self.add_to_config(config_data, RestConfiguration(name="rest"))

        if 'all' in args.clients or 'facebook' in args.clients:
            self.add_to_config(config_data, FacebookConfiguration(), args.defaults)

        if 'all' in args.clients or 'kik' in args.clients:
            self.add_to_config(config_data, KikConfiguration(), args.defaults)

        if 'all' in args.clients or 'line' in args.clients:
            self.add_to_config(config_data, LineConfiguration(), args.defaults)

        if 'all' in args.clients or 'twilio' in args.clients:
            self.add_to_config(config_data, TwilioConfiguration(), args.defaults)

        if 'all' in args.clients or 'viber' in args.clients:
            self.add_to_config(config_data, ViberConfiguration(), args.defaults)

        if 'all' in args.clients or 'sanic' in args.clients:
            self.add_to_config(config_data, SanicRestConfiguration(name="sanic"))

        client_config = ConsoleConfiguration()

        bot_config = client_config.configurations[0]

        self.add_to_config(config_data, bot_config, args.defaults)

        brain_config = bot_config.configurations[0]

        self.add_to_config(config_data, brain_config, args.defaults)

        self.write_yaml(args.file, config_data)

    def write_yaml(self, filename, data):
        outputLog(self, "Writing new config file to [%s]" % filename)
        try:
            with open(filename, 'w') as outfile:
                yaml.dump(data, outfile, default_flow_style=False)

        except Exception as excep:
            outputLog(self, "Failed to write new config file [%s] - [%s]" % (filename, str(excep)))

    @staticmethod
    def create_arguments():
        parser = argparse.ArgumentParser(description='Program-Y Configuration Writer')

        parser.add_argument('-f', '--file', default="config.yaml", help="Name of file to create")
        parser.add_argument('-c', '--clients', nargs='+',
                            help="Name of client config to create, use multiple times or all")
        parser.add_argument('-d', '--defaults', action='store_true',
                            help="Create all config settings with default values")

        return parser

    @staticmethod
    def run(args=None):
        parser = ConfigurationWriter.create_arguments()
        try:
            app = ConfigurationWriter()

            if args is None:
                app.execute(parser.parse_args())

            else:
                app.execute(args)

        except Exception as excep:
            outputLog(None, "Error: [%s]" % str(excep))
            parser.print_help()


if __name__ == '__main__':      # pragma: no cover
    ConfigurationWriter.run()   # pragma: no cover



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from abc import ABC
from abc import abstractmethod
from programy.utils.substitutions.substitues import Substitutions


class BaseConfigurationFile(ABC):

    @abstractmethod
    def load_from_text(self, text, client_configuration, bot_root, subs: Substitutions = None):
        """
        Never Implemented
        """
        raise NotImplementedError()  # pragma: no cover

    @abstractmethod
    def load_from_file(self, filename, client_configuration, bot_root, subs: Substitutions = None):
        """
        Never Implemented
        """
        raise NotImplementedError()  # pragma: no cover

    @abstractmethod
    def get_section(self, section_name, parent_section=None):
        """
        Never Implemented
        """
        raise NotImplementedError()  # pragma: no cover

    @abstractmethod
    def get_keys(self, section):
        """
        Never Implemented
        """
        raise NotImplementedError()  # pragma: no cover

    @abstractmethod
    def get_child_section_keys(self, child_section_name, parent_section):
        """
        Never Implemented
        """
        raise NotImplementedError()  # pragma: no cover

    def _replace_subs(self, subs: Substitutions, value):
        if subs is not None:
            return subs.replace(value)

        return value

    @abstractmethod
    def get_option(self, section, option_name, missing_value=None, subs: Substitutions = None):
        """
        Never Implemented
        """
        raise NotImplementedError()  # pragma: no cover

    @abstractmethod
    def get_bool_option(self, section, option_name, missing_value=False, subs: Substitutions = None):
        """
        Never Implemented
        """
        raise NotImplementedError()  # pragma: no cover

    @abstractmethod
    def get_int_option(self, section, option_name, missing_value=0, subs: Substitutions = None):
        """
        Never Implemented
        """
        raise NotImplementedError()  # pragma: no cover

    @abstractmethod
    def get_multi_option(self, section, option_name, missing_value=None, subs: Substitutions = None):
        """
        Never Implemented
        """
        raise NotImplementedError()  # pragma: no cover

    @abstractmethod
    def get_multi_file_option(self, section, option_name, bot_root, missing_value=None, subs: Substitutions = None):
        """
        Never Implemented
        """
        raise NotImplementedError()  # pragma: no cover

    def convert_to_bool(self, value):
        if value.upper() == 'TRUE':
            return True
        elif value.upper() == 'FALSE':
            return False
        else:
            raise Exception("Invalid boolean config value [%s]" % value)

    def convert_to_int(self, value):
        if value.isdigit():
            return int(value)
        else:
            raise Exception("Invalid integer config value [%s]" % value)



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import datetime
import locale
from programy.utils.logging.ylogger import YLogger
from programy.parser.exceptions import ParserException
from programy.parser.template.nodes.attrib import TemplateAttribNode
from programy.parser.template.nodes.word import TemplateWordNode
from programy.parser.template.nodes.base import TemplateNode


class TemplateDateNode(TemplateAttribNode):

    def __init__(self, date_format=None, locale=None):
        TemplateAttribNode.__init__(self)

        if date_format is None:
            self._format = TemplateWordNode("%c")
        else:
            self._format = TemplateWordNode(date_format)

        self._locale = None
        if locale is not None:
            self._locale = TemplateWordNode(locale)

    @property
    def date_format(self):
        return self._format

    @date_format.setter
    def date_format(self, fmt):
        if isinstance(fmt, TemplateNode):
            self._format = fmt
        else:
            self._format = TemplateWordNode(str(fmt))

    @property
    def locale(self):
        return self._locale

    @locale.setter
    def locale(self, lcl):
        if isinstance(lcl, TemplateNode):
            self._locale = lcl
        else:
            self._locale = TemplateWordNode(str(lcl))

    def _set_locate(self, client_context):
        local_locale = self._locale.resolve_to_string(client_context)
        locale.setlocale(locale.LC_TIME, local_locale)
        return local_locale

    def resolve_to_string(self, client_context):
        time_now = datetime.datetime.now()

        local_time = None
        if self._locale is not None:
            local_time = locale.setlocale(locale.LC_TIME)

        local_locale = ""
        resolved = ""
        try:
            if self._locale is not None:
                local_locale = self._set_locate(client_context)

            resolved_format = self._format.resolve_to_string(client_context)
            resolved = time_now.strftime(resolved_format)

        except Exception as exep:
            YLogger.exception(self, "Failed to set locale to [%s]", exep, local_locale)

        finally:
            if local_time is not None:
                locale.setlocale(locale.LC_TIME, local_time)

        YLogger.debug(client_context, "[%s] resolved to [%s]", self.to_string(), resolved)
        return resolved

    def to_string(self):
        if self._locale is None:
            return "[DATE format=%s]" % self._format.to_string()
        else:
            return "[DATE format=%s locale=%s]" % (self._format.to_string(), self._locale.to_string())

    def set_attrib(self, attrib_name, attrib_value):

        if attrib_name == 'format':
            if isinstance(attrib_value, TemplateNode):
                self._format = attrib_value
            else:
                self._format = TemplateWordNode(attrib_value)

        elif attrib_name == 'locale':
            if isinstance(attrib_value, TemplateNode):
                self._locale = attrib_value
            else:
                self._locale = TemplateWordNode(attrib_value)

        else:
            raise ParserException("Invalid attribute name %s for this node" % (attrib_name))

    def to_xml(self, client_context):
        if self._locale is None:
            xml = '<date format="%s" >' % self._format.to_xml(client_context)
        else:
            xml = '<date format="%s" locale="%s">' % (self._format.to_xml(client_context),
                                                      self._locale.to_xml(client_context))

        xml += self.children_to_xml(client_context)
        xml += "</date>"
        return xml

    #######################################################################################################
    # DATE_ATTRIBUTES ::== (format="LISP_DATE_FORMAT") | (jformat="JAVA DATE FORMAT")
    # DATE_ATTRIBUTE_TAG ::== <format>TEMPLATE_EXPRESSION</format> | <jformat>TEMPLATE_EXPRESSION</jformat>
    # DATE_EXPRESSION ::== <date( DATE_ATTRIBUTES)*/> | <date>(DATE_ATTRIBUTE_TAG)</date>
    # Pandorabots supports three extension attributes to the date element in templates:
    #     	locale
    #       format
    #       timezone

    def parse_expression(self, graph, expression):
        self._parse_node_with_attribs(graph, expression, [["format", "%c"], ["locale", None]])



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import json
from flask import Flask, request
from programy.utils.logging.ylogger import YLogger
from programy.clients.restful.flask.client import FlaskRestBotClient
from programy.clients.restful.flask.google.config import GoogleConfiguration
from programy.utils.console.console import outputLog


class GoogleBotClient(FlaskRestBotClient):

    def __init__(self, argument_parser=None):
        FlaskRestBotClient.__init__(self, 'google', argument_parser)

        YLogger.debug(self, "Google Client is running....")

        outputLog(self, "Google Client loaded")

    def _to_json(self, data):
        return json.dumps(data)

    def get_client_configuration(self):
        return GoogleConfiguration()

    def _ask_question(self, client_context, question):
        reply = ""
        try:
            self._questions += 1
            response = client_context.bot.ask_question(client_context, question, responselogger=self)
            reply = self.renderer.render(client_context, response)

        except Exception as e:
            YLogger.exception(client_context, "Error getting reply from bot", e)

        return reply

    def _get_reply_from_bot(self, client_context, text, srai):
        if srai is not None:
            reply = self._ask_question(client_context, srai)
            if reply is None or reply == "":
                reply = text
        else:
            reply = text

        return reply

    def _handle_launch_intent(self, client_context):
        response = self._get_reply_from_bot(client_context, self.configuration.client_configuration.launch_text,
                                            self.configuration.client_configuration.launch_srai)
        return self._to_json({"fulfillmentText": response})

    def _handle_quit_intent(self, client_context):
        response = self._get_reply_from_bot(client_context, self.configuration.client_configuration.quit_text,
                                            self.configuration.client_configuration.quit_srai)
        return self._to_json({"fulfillmentText": response})

    def _handle_help_intent(self, client_context):
        response = self._get_reply_from_bot(client_context, self.configuration.client_configuration.help_text,
                                            self.configuration.client_configuration.help_srai)
        return self._to_json({"fulfillmentText": response})

    def _handle_error(self, client_context):
        response = self._get_reply_from_bot(client_context, self.configuration.client_configuration.error_text,
                                            self.configuration.client_configuration.error_srai)
        return self._to_json({"fulfillmentText": response})

    def _handle_query_intent(self, client_context, query_result):

        if 'queryText' not in query_result:
            raise Exception("Invalid http queryResult, queryText missing!")
        query_text = query_result['queryText']

        response = self._ask_question(client_context, query_text)

        return self._to_json({"fulfillmentText": response})

    def _get_userid(self, skill_data):

        if 'session' in skill_data:
            return skill_data['session']

        return 'google'

    def receive_message(self, http_request):

        skill_data = json.loads(http_request.data)
        #output(json.dumps(skill_data, indent=4))

        if 'queryResult' not in skill_data:
            raise Exception("Invalid http request, queryResult missing!")
        queryResult = skill_data['queryResult']

        if 'intent' not in queryResult:
            raise Exception("Invalid queryResult, intent missing!")
        intent = queryResult['intent']

        if 'displayName' not in intent:
            raise Exception("Invalid intent, intent displayName!")

        userid = self._get_userid(skill_data)

        client_context = self.create_client_context(userid)

        intent_name = intent['displayName']

        try:
            if intent_name == 'Launch Intent':
                return self._handle_launch_intent(client_context)

            elif intent_name == 'Query Intent':
                return self._handle_query_intent(client_context, queryResult)

            elif intent_name == 'Quit Intent':
                return self._handle_quit_intent(client_context)

            elif intent_name == 'Help Intent':
                return self._handle_help_intent(client_context)

            else:
                raise Exception("Invalid intent name [%s]!" % intent_name)

        except Exception as e:
            YLogger.exception(client_context, "Unknown/Unhandled intent [%s]", e, intent_name)
            return self._handle_error(client_context)


if __name__ == "__main__":

    outputLog(None, "Initiating Google Client...")

    GOOGLE_CLIENT = GoogleBotClient()

    APP = Flask(__name__)

    @APP.route("/api/google/v1.0/ask", methods=['GET', 'POST'])
    def receive_message():
        try:
            return GOOGLE_CLIENT.receive_message(request)

        except Exception as e:
            outputLog(None, "Failure receiving message", e)
            YLogger.exception(None, "Google Error", e)

    GOOGLE_CLIENT.run(APP)



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions
of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from programy.utils.logging.ylogger import YLogger
from programy.utils.weather.metoffice import MetOffice
from programy.utils.weather.metoffice import MetOffice5DayForecast
from programy.utils.weather.metoffice import MetOffice24HourForecast
from programy.utils.geo.google import GoogleMaps
from programy.extensions.base import Extension


class WeatherExtension(Extension):

    def get_geo_locator(self):
        return GoogleMaps()

    def get_met_office(self):
        return MetOffice()

    def get_from_date(self):
        return None

    # WEATHER [OBSERVATION|FORECAST3|FORECAST24] LOCATION * WHEN *

    # execute() is the interface that is called from the <extension> tag in the AIML
    def execute(self, client_context, data):

        try:
            fromdate = self.get_from_date()

            splits = data.split()
            if len(splits) != 5:
                YLogger.debug(client_context, "Weather - Not enough paramters passed, [%d] expected 5", len(splits))
                return None

            obvtype = splits[0]
            if obvtype not in ['OBSERVATION', 'FORECAST5DAY', 'FORECAST24HOUR']:
                YLogger.debug(client_context, "Weather - Type not understood [%s]", obvtype)
                return None

            if splits[1] == 'LOCATION':
                postcode = splits[2]
            else:
                YLogger.debug(client_context, "Weather - LOCATION missing")
                return None

            if splits[3] == 'WHEN':
                when = splits[4]
            else:
                YLogger.debug(client_context, "Weather - WHEN missing")
                return None

            if obvtype == 'OBSERVATION':
                result = self.current_observation(client_context, postcode)

            elif obvtype == 'FORECAST5DAY':
                result =  self.five_day_forecast(client_context, postcode, when, fromdate=fromdate)

            else: # obvtype == 'FORECAST24HOUR':
                result = self.twentyfour_hour_forecast(client_context, postcode, when, fromdate=fromdate)

            return result

        except Exception as error:
            YLogger.exception(client_context, "Failed to execute weather extension", error)

        return "UNAVAILABLE"

    def current_observation(self, context, postcode):
        YLogger.debug(context, "Getting weather observation for [%s]", postcode)

        googlemaps = self.get_geo_locator()
        latlng = googlemaps.get_latlong_for_location(postcode)

        met_office = self.get_met_office()

        observation = met_office.current_observation(latlng.latitude, latlng.longitude)
        if observation is not None:
            return observation.get_latest_observation().to_program_y_text()
        else:
            return "UNAVAILABLE"

    def twentyfour_hour_forecast(self, context, postcode, when, fromdate=None):
        YLogger.debug(context, "Getting 24 hour weather forecast for [%s] at time [%s]", postcode, when)

        googlemaps = self.get_geo_locator()
        latlng = googlemaps.get_latlong_for_location(postcode)

        met_office = self.get_met_office()

        forecast = met_office.twentyfour_hour_forecast(latlng.latitude, latlng.longitude)
        if forecast is not None:

            datapoint = forecast.get_forecast_for_n_hours_ahead(int(when), fromdate=fromdate)
            if datapoint is not None:
                return datapoint

            datapoint = forecast.get_latest_forecast().to_program_y_text()
            if datapoint is not None:
                return datapoint

        return "UNAVAILABLE"

    def five_day_forecast(self, context, postcode, when, fromdate=None):
        YLogger.debug(context, "Getting 5 day forecast for [%s] at time [%s]", postcode, when)

        googlemaps = self.get_geo_locator()
        latlng = googlemaps.get_latlong_for_location(postcode)

        met_office = self.get_met_office()

        forecast = met_office.five_day_forecast(latlng.latitude, latlng.longitude)
        if forecast is not None:

            datapoint = forecast.get_forecast_for_n_days_ahead(int(when), fromdate=fromdate)
            if datapoint is not None:
                return datapoint

            datapoint = forecast.get_latest_forecast().to_program_y_text()
            if datapoint is not None:
                return datapoint

        return "UNAVAILABLE"



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions
of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from abc import ABC
from abc import abstractmethod
from programy.utils.logging.ylogger import YLogger
from programy.utils.classes.loader import ClassLoader
from programy.activate import Activatable


class SpellingChecker(Activatable, ABC):

    def __init__(self, spelling_config=None):
        Activatable.__init__(self)
        self.spelling_config = spelling_config

    def initialise(self, storage_factory):
        pass

    @abstractmethod
    def correct(self, phrase):
        raise NotImplementedError()  # pragma: no cover

    @staticmethod
    def initiate_spellchecker(spelling_config, storage_factory):
        if spelling_config.classname is not None:
            try:
                YLogger.info(None, "Loading spelling checker from class [%s]", spelling_config.classname)
                spell_class = ClassLoader.instantiate_class(spelling_config.classname)
                spell_checker = spell_class(spelling_config)
                spell_checker.initialise(storage_factory)
                return spell_checker
            except Exception as excep:
                YLogger.exception(None, "Failed to initiate spellcheker", excep)
        else:
            YLogger.warning(None, "No configuration setting for spelling checker!")

        return None
    
    def check_spelling_before(self, client_context, each_sentence):
        if self.is_active():
            if self.spelling_config.check_before is True:
                text = each_sentence.text(client_context)
                corrected = self.correct(text)
                YLogger.debug(client_context, "Spell Checker corrected [%s] to [%s]", text, corrected)
                each_sentence.replace_words(client_context, corrected)

        else:
            YLogger.debug(client_context, "Spelling is switched off.")

    def check_spelling_and_retry(self, client_context, each_sentence):
        if self.is_active():
            if self.spelling_config.check_and_retry is True:
                text = each_sentence.text(client_context)
                corrected = self.correct(text)
                YLogger.debug(client_context, "Spell Checker corrected [%s] to [%s]", text, corrected)
                each_sentence.replace_words(client_context, corrected)
                response = client_context.brain.ask_question(client_context, each_sentence)
                return response

        else:
            YLogger.debug(client_context, "Spelling is switched off.")

        return None



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from flask import Flask, jsonify, request, make_response, abort, Response

from programy.clients.restful.client import RestBotClient
from programy.utils.console.console import outputLog


class FlaskRestBotClient(RestBotClient):

    def __init__(self, botid, argument_parser=None):
        RestBotClient.__init__(self, botid, argument_parser)
        self.initialise()

    def _render_callback(self):
        return False

    def server_abort(self, message, status_code):
        abort(Response(message), status_code)

    def make_response_v1(self, data, status_code):
        return make_response(jsonify(data, status_code))

    def make_response_v2(self, data, status_code):
        return make_response(jsonify(data), status_code)

    def make_response_other(self, data, status_code):
        return make_response(data, status_code)

    def create_response(self, response_data, status_code, version=1.0):
        if self.configuration.client_configuration.debug is True:
            self.dump_request(response_data)

        if version == 1.0:
            return self.make_response_v1(response_data, status_code)
        elif version == 2.0:
            return self.make_response_v2(response_data, status_code)

        return self.make_response_other('Invalid API version', 400)

    def run(self, app=None):

        outputLog(self, "%s Client running on http://%s:%s" % (self.id, self.configuration.client_configuration.host,
                                                               self.configuration.client_configuration.port))

        self.startup()

        if self.configuration.client_configuration.debug is True:
            outputLog(self, "%s Client running in debug mode" % self.id)

        if self.configuration.client_configuration.ssl_cert_file is not None and \
                self.configuration.client_configuration.ssl_key_file is not None:
            context = (self.configuration.client_configuration.ssl_cert_file,
                       self.configuration.client_configuration.ssl_key_file)

            outputLog(self, "%s Client running in https mode" % self.id)
            app.run(host=self.configuration.client_configuration.host,
                    port=self.configuration.client_configuration.port,
                    debug=self.configuration.client_configuration.debug,
                    ssl_context=context)
        else:
            outputLog(self, "%s Client running in http mode, careful now !" % self.id)
            app.run(host=self.configuration.client_configuration.host,
                    port=self.configuration.client_configuration.port,
                    debug=self.configuration.client_configuration.debug)

        self.shutdown()


if __name__ == '__main__':
    REST_CLIENT = None

    outputLog(None, "Initiating Flask REST Service...")
    APP = Flask(__name__)


    @APP.route('/api/rest/v1.0/ask', methods=['GET', 'POST'])
    def ask_v1_0():
        response_data, status = REST_CLIENT.process_request(request, version=1.0)
        return REST_CLIENT.create_response(response_data, status, version=1.0)


    @APP.route('/api/rest/v2.0/ask', methods=['GET', 'POST'])
    def ask_v2_0():
        response_data, status = REST_CLIENT.process_request(request, version=2.0)
        return REST_CLIENT.create_response(response_data, status, version=2.0)


    outputLog(None, "Loading, please wait...")
    REST_CLIENT = FlaskRestBotClient("flask")
    REST_CLIENT.run(APP)



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

class StorageEngineException(Exception):

    def __init__(self, msg):
        Exception.__init__(self, msg)


class StorageEngine:

    def __init__(self, configuration):
        self._configuration = configuration

    @property
    def configuration(self):
        return self._configuration

    def user_store(self):
        raise StorageEngineException("Engine does not support user storage")

    def linked_account_store(self):
        raise StorageEngineException("Engine does not support linked account storage")

    def link_store(self):
        raise StorageEngineException("Engine does not support link storage")

    def category_store(self):
        raise StorageEngineException("Engine does not support category storage")

    def errors_store(self):
        raise StorageEngineException("Engine does not support errors storage")

    def duplicates_store(self):
        raise StorageEngineException("Engine does not support duplicates storage")

    def learnf_store(self):
        raise StorageEngineException("Engine does not support learnf storage")

    def conversation_store(self):
        raise StorageEngineException("Engine does not support conversation storage")

    def sets_store(self):
        raise StorageEngineException("Engine does not support sets storage")

    def maps_store(self):
        raise StorageEngineException("Engine does not support maps storage")

    def rdf_store(self):
        raise StorageEngineException("Engine does not support rdf storage")

    def denormal_store(self):
        raise StorageEngineException("Engine does not support denormal storage")

    def normal_store(self):
        raise StorageEngineException("Engine does not support normal storage")

    def gender_store(self):
        raise StorageEngineException("Engine does not support gender storage")

    def person_store(self):
        raise StorageEngineException("Engine does not support person storage")

    def person2_store(self):
        raise StorageEngineException("Engine does not support person2 storage")

    def regex_store(self):
        raise StorageEngineException("Engine does not support regex storage")

    def property_store(self):
        raise StorageEngineException("Engine does not support property storage")

    def defaults_store(self):
        raise StorageEngineException("Engine does not support variables storage")

    def variables_store(self):
        raise StorageEngineException("Engine does not support variables storage")

    def twitter_store(self):
        raise StorageEngineException("Engine does not support twitter storage")

    def spelling_store(self):
        raise StorageEngineException("Engine does not support spelling storage")

    def license_store(self):
        raise StorageEngineException("Engine does not support license storage")

    def pattern_nodes_store(self):
        raise StorageEngineException("Engine does not support pattern nodes storage")

    def template_nodes_store(self):
        raise StorageEngineException("Engine does not support template nodes storage")

    def binaries_store(self):
        raise StorageEngineException("Engine does not support binaries storage")

    def braintree_store(self):
        raise StorageEngineException("Engine does not support braintree storage")

    def preprocessors_store(self):
        raise StorageEngineException("Engine does not support preprocessor storage")

    def postprocessors_store(self):
        raise StorageEngineException("Engine does not support postprocessor storage")

    def postquestionprocessors_store(self):
        raise StorageEngineException("Engine does not support postprocessor storage")

    def usergroups_store(self):
        raise StorageEngineException("Engine does not support user groups storage")

    def triggers_store(self):
        raise StorageEngineException("Engine does not support trigger storage")

    def oobs_store(self):
        raise StorageEngineException("Engine does not support oobs storage")



import smtplib
import mimetypes
from email import encoders
from email.mime.audio import MIMEAudio
from email.mime.base import MIMEBase
from email.mime.image import MIMEImage
from email.mime.text import MIMEText
from email.mime.multipart import MIMEMultipart
from programy.utils.logging.ylogger import YLogger
from programy.utils.email.config import EmailConfiguration


class EmailSender:

    def __init__(self, config: EmailConfiguration):
        self._config = config
        self._attachments = []

    def _add_mine_attachments(self, msg, attachments):
        for attachment in attachments:
            if isinstance(attachment, tuple):
                self._add_attachement(msg, attachment[0], attachment[1], attachment[2])
            else:
                self._add_attachement(msg, attachment)

    def _guess_mime_type(self, path):
        return mimetypes.guess_type(path)

    def _split_ctype(self, ctype):
        return ctype.split('/', 1)

    def _get_ctype_and_attachment(self, path, encoding):

        ctype, attachment_encoding = self._guess_mime_type(path)

        if ctype is None:
            # No guess could be made, or the file is encoded (compressed), so
            # use a generic bag-of-bits type.
            ctype = 'application/octet-stream'

        if attachment_encoding is None:
            if encoding is not None:
                attachment_encoding = encoding

            else:
                # No guess could be made, so we default to utf-8
                attachment_encoding = "utf-8"

        return ctype, attachment_encoding

    def _attach_text(self, msg, path, encoding, subtype):
        with open(path, encoding=encoding) as fp:
            # Note: we should handle calculating the charset
            attach = MIMEText(fp.read(), _subtype=subtype)
            msg.attach(attach)

    def _attach_image(self, msg, path, subtype):
        with open(path, 'rb') as fp:
            attach = MIMEImage(fp.read(), _subtype=subtype)
            msg.attach(attach)

    def _attach_audio(self, msg, path, subtype):
        with open(path, 'rb') as fp:
            attach = MIMEAudio(fp.read(), _subtype=subtype)
            msg.attach(attach)

    def _attach_binary(self, msg, path, mimetype, subtype):
        # No specific mime type, so we shoot for a binary file
        with open(path, 'rb') as fp:
            attach = MIMEBase(mimetype, subtype)
            attach.set_payload(fp.read())
            encoders.encode_base64(attach)
            msg.attach(attach)

    def _add_attachement(self, msg, path, ctype=None, encoding=None):

        if ctype is None:
            ctype, encoding = self._get_ctype_and_attachment(path, encoding)

        if encoding is None:
            encoding = "utf-8"

        mimetype, subtype = self._split_ctype(ctype)

        if mimetype == 'text':
            self._attach_text(msg, path, encoding, subtype)

        elif mimetype == 'image':
            self._attach_image(msg, path, subtype)

        elif mimetype == 'audio':
            self._attach_audio(msg, path, subtype)

        else:
            self._attach_binary(msg, path, mimetype, subtype)

    def _smtp_server(self, host, port):
        return smtplib.SMTP(host, port)     # pragma: no cover

    def _send_message(self, host, port, username, password, msg):

        YLogger.info(self, "Email sender starting")
        server = self._smtp_server(host, port)
        server.ehlo()
        server.starttls()
        YLogger.info(self, "Email sender logging in")
        server.login(username, password)
        YLogger.info(self, "Email sender sending")
        result = server.send_message(msg)
        YLogger.info(self, "Email sender quiting")
        server.quit()
        return result

    def send(self, to, subject, message, attachments=None):

        try:
            if attachments:
                YLogger.info(self, "Email sender adding mime attachment")
                msg = MIMEMultipart()
                msg.attach(MIMEText(message))
                self._add_mine_attachments(msg, attachments)
            else:
                msg = MIMEText(message)

            msg['Subject'] = subject
            msg['From'] = self._config.from_addr
            msg['To'] = to

            result = self._send_message(self._config.host,
                                        self._config.port,
                                        self._config.username,
                                        self._config.password,
                                        msg)

            if result:
                for email, error in result.items():
                    YLogger.error(None, "Email send failed: [%d] - [%s]", email, error)

        except Exception as e:
            YLogger.exception(self, "Email sender failed", e)



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import re
from programy.utils.logging.ylogger import YLogger
from programy.utils.text.text import TextUtils
from programy.dialog.question import Question


class Conversation():

    def __init__(self, client_context):
        self._client_context = client_context
        self._questions = []
        self._max_histories = client_context.bot.configuration.conversations.max_histories
        self._properties = {'topic': client_context.bot.configuration.conversations.initial_topic}

    @property
    def questions(self):
        return self._questions

    @property
    def max_histories(self):
        return self._max_histories

    @max_histories.setter
    def max_histories(self, histories):
        self._max_histories = histories

    @property
    def properties(self):
        return self._properties

    def has_current_question(self):
        return bool(self._questions)

    def current_question(self):
        if self._questions:
            return self._questions[-1]
        raise Exception("Invalid question index")

    def previous_nth_question(self, num: int):
        if len(self._questions) < num:
            raise Exception("Num question array violation !")
        previous = -1 - num
        return self._questions[previous]

    def set_property(self, name: str, value: str):
        if name == 'topic':
            if value == "":
                value = '*'
        self._properties[name] = value

    def property(self, name: str):
        if name in self._properties:
            return self._properties[name]
        return None

    def record_dialog(self, question: Question):
        if len(self._questions) == self._max_histories:
            YLogger.info(self, "Conversation history at max [%d], removing oldest", self._max_histories)
            self._questions.remove(self._questions[0])
        self._questions.append(question)

    def pop_dialog(self):
        if self._questions:
            self._questions.pop()

    def load_initial_variables(self, variables_collection):
        for pair in variables_collection.pairs:
            YLogger.debug(self, "Setting variable [%s] = [%s]", pair[0], pair[1])
            self._properties[pair[0]] = pair[1]

    def get_topic_pattern(self, client_context):
        topic_pattern = self.property("topic")

        if topic_pattern is None:
            YLogger.info(client_context, "No Topic pattern default to [*]")
            topic_pattern = "*"
        else:
            YLogger.info(client_context, "Topic pattern = [%s]", topic_pattern)

        return topic_pattern

    def parse_last_sentences_from_response(self, response):

        # If the response contains punctuation such as "Hello. There" then THAT is none
        response = re.sub(r'<\s*br\s*/>\s*', ".", response)
        response = re.sub(r'<br></br>*', ".", response)
        sentences = response.split(".")
        sentences = [x for x in sentences if x]
        if sentences:
            last_sentence = sentences[-1]
            that_pattern = TextUtils.strip_all_punctuation(last_sentence)
            that_pattern = that_pattern.strip()
        else:
            that_pattern = "*"

        return that_pattern

    def get_that_pattern(self, client_context, srai=False):
        try:
            that_question = None
            if srai is False:
                that_question = self.previous_nth_question(1)
            else:
                if len(self._questions) > 2:
                    for question in reversed(self._questions[:-2]):
                        if question.srai is False and question.has_response():
                            that_question = question
                            break

            if that_question is not None:
                that_sentence = that_question.current_sentence()
            else:
                that_sentence = None

            # If the last response was valid, i.e not none and not empty string, then use
            # that as the that_pattern, otherwise we default to '*' as pattern
            if that_sentence is not None and that_sentence.response is not None and that_sentence.response != '':
                that_pattern = self.parse_last_sentences_from_response(that_sentence.response)
                YLogger.info(client_context, "That pattern = [%s]", that_pattern)
            else:
                YLogger.info(client_context, "That pattern, no response, default to [*]")
                that_pattern = "*"

        except Exception as excep:
            YLogger.exception_nostack(client_context, "No That pattern default to [*]", excep)
            that_pattern = "*"

        return that_pattern

    def recalculate_sentiment_score(self, client_context):
        for question in self._questions:
            question.recalculate_sentinment_score(client_context)

    def calculate_sentiment_score(self):

        positivity = 0.00
        subjectivity = 0.00

        count = 0
        for question in self._questions:
            q_positivity, q_subjectivity = question.calculate_sentinment_score()

            positivity += q_positivity
            subjectivity += q_subjectivity

            count += 1

        if count > 0:
            positivity /= count
            subjectivity /= count
        else:
            subjectivity = 0.5

        return positivity, subjectivity

    def save_sentiment(self):
        positivity, subjectivity = self.calculate_sentiment_score()
        self._properties['positivity'] = str(positivity)
        self._properties['subjectivity'] = str(subjectivity)

    def to_json(self):
        json_data = {
            'client_context': self._client_context.to_json(),
            'questions': [],
            'max_histories': self._max_histories,
            'properties': self._properties
        }

        for question in self.questions:
            json_data["questions"].append(question.to_json())

        return json_data

    @staticmethod
    def from_json(client_context, json_data):
        conversation = Conversation(client_context)
        conversation.create_from_json(json_data)
        return conversation

    def create_from_json(self, json_data):

        for key, value in json_data['properties'].items():
            self._properties[key] = value

        for json_question in json_data['questions']:
            self._questions.append(Question.from_json(self._client_context, json_question))

        self.recalculate_sentiment_score(self._client_context)



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import json
from programy.utils.logging.ylogger import YLogger
from programy.parser.template.nodes.base import TemplateNode
from programy.parser.exceptions import ParserException
from programy.utils.text.text import TextUtils


class TemplateGetNode(TemplateNode):

    def __init__(self):
        TemplateNode.__init__(self)
        self._name = None
        self._local = False
        self._tuples = None

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        self._name = name

    @property
    def local(self):
        return self._local

    @local.setter
    def local(self, local):
        self._local = local

    @property
    def tuples(self):
        return self._tuples

    @tuples.setter
    def tuples(self, tuples):
        self._tuples = tuples

    @staticmethod
    def get_default_value(client_context):
        value = client_context.bot.brain.properties.property("default_get")
        if value is None:
            YLogger.error(None, "No property defined for default_get, checking defaults")

            value = client_context.bot.brain.configuration.defaults.default_get
            if value is None:
                YLogger.error(None, "No value defined for default default_get, returning 'unknown'")
                value = "unknown"

        return value

    @staticmethod
    def get_property_value(client_context, local, name):

        conversation = client_context.bot.get_conversation(client_context)

        value = None
        if local is True:

            if conversation.has_current_question():
                value = conversation.current_question().property(name)

        else:

            if name is not None and client_context.brain.dynamics.is_dynamic_var(name) is True:
                value = client_context.brain.dynamics.dynamic_var(client_context, name)
            else:
                value = conversation.property(name)

        if value is None:
            YLogger.error(client_context, "No property for [%s]", name)
            value = TemplateGetNode.get_default_value(client_context)

        return value

    def resolve_variable(self, client_context):
        name = self.name.resolve(client_context)
        value = TemplateGetNode.get_property_value(client_context, self.local, name)
        if self.local:
            YLogger.debug(client_context, "[%s] resolved to local: [%s] <= [%s]", self.to_string(), name, value)
        else:
            YLogger.debug(client_context, "[%s] resolved to global: [%s] <= [%s]", self.to_string(), name, value)
        return value

    @staticmethod
    def decode_tuples(tuples):
        if isinstance(tuples, str):
            return json.loads(tuples)
        else:
            return tuples

    def _get_tuples(self, client_context):
        raw_tuples = self._tuples.resolve(client_context)
        try:
            return TemplateGetNode.decode_tuples(raw_tuples)

        except Exception as excep:
            YLogger.exception_nostack(self, "Failed to decode tuples, returning []", excep)
            return []

    def resolve_tuple(self, client_context):

        variables = None
        if self._name:
            variables = self._name.resolve(client_context).split(" ")

        tuples = self._get_tuples(client_context)

        resolved = ""
        if variables:  # If we are asking for variables, pull out the vars
            for atuple in tuples:
                if isinstance(atuple[0], list) is True:
                    for pair in atuple:
                        for var in variables:
                            if pair[0] == var:
                                resolved += pair[1]
                                resolved += " "
                else:
                    for var in variables:
                        if atuple[0] == var:
                            resolved += atuple[1]
                            resolved += " "

        else:
            for atuple in tuples:
                resolved += atuple[0][1]
                resolved += " "
                resolved += atuple[1][1]
                resolved += " "
                resolved += atuple[2][1]
                resolved += " "

        YLogger.debug(client_context, "[%s] resolved to [%s]", self.to_string(), resolved)
        value = resolved.strip()

        if value is None or value == "":
            YLogger.error(client_context, "No value for tuple, returning default_get")
            value = TemplateGetNode.get_default_value(client_context)

        return value

    def resolve_to_string(self, client_context):
        if self._tuples is None:
            value = self.resolve_variable(client_context)
        else:
            value = self.resolve_tuple(client_context)
        return value

    def to_string(self):
        if self.tuples is None:
            if self.name is None:
                name = "None"
            else:
                name = self.name.to_string()
            return "[GET [%s] - %s]" % ("Local" if self.local else "Global", name)
        else:
            if self.name:
                return "[GET [Tuples] - (%s)]" % self.name.to_string()
            else:
                return "[GET [Tuples]"

    def to_xml(self, client_context):
        if self.tuples is None:
            xml = "<get"
            if self.local:
                xml += ' var="%s"' % self.name.resolve(client_context)
            else:
                xml += ' name="%s"' % self.name.resolve(client_context)
            xml += " />"
        else:
            xml = "<get"
            xml += ' var="%s"' % self.name.resolve(client_context)
            xml += " >"
            xml += self.tuples.to_xml(client_context)
            xml += "</get>"
        return xml

    # ######################################################################################################
    # GET_PREDICATE_EXPRESSION ::==
    # <get name="WORD"/> |
    # <get><name>TEMPLATE_EXPRESSION</name></get> |
    # <get var=”WORD”> |
    # <get><var>WORD</var></get>

    def parse_expression(self, graph, expression):

        name_found = False
        var_found = False

        if 'name' in expression.attrib:
            self.name = self.parse_attrib_value_as_word_node(graph, expression, 'name')
            self.local = False
            name_found = True

        if 'var' in expression.attrib:
            self.name = self.parse_attrib_value_as_word_node(graph, expression, 'var')
            self.local = True
            var_found = True

        for child in expression:
            tag_name = TextUtils.tag_from_text(child.tag)

            if tag_name == 'name':
                self.name = self.parse_children_as_word_node(graph, child)
                self.local = False
                name_found = True

            elif tag_name == 'var':
                self.name = self.parse_children_as_word_node(graph, child)
                self.local = True
                var_found = True

            elif tag_name == "tuple":
                self._tuples = self.parse_children_as_word_node(graph, child)

        if name_found is False and var_found is False:
            raise ParserException("Invalid get, missing either name or var", xml_element=expression)

        if name_found is True and var_found is True:
            raise ParserException("Get node has both name AND var values", xml_element=expression)



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in all copies or substantial portions
of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

import os
import logging.config
import yaml

from programy.utils.logging.ylogger import YLogger

from programy.config.file.factory import ConfigurationFactory
from programy.clients.args import CommandLineClientArguments
from programy.config.programy import ProgramyConfiguration
from programy.context import ClientContext
from programy.utils.license.keys import LicenseKeys
from programy.utils.substitutions.substitues import Substitutions
from programy.scheduling.scheduler import ProgramyScheduler
from programy.clients.render.text import TextRenderer
from programy.utils.classes.loader import ClassLoader
from programy.storage.factory import StorageFactory
from programy.utils.email.sender import EmailSender
from programy.triggers.manager import TriggerManager
from programy.triggers.system import SystemTriggers
from programy.clients.ping.responder import PingResponder
from programy.clients.botfactory import BotFactory
from programy.clients.response import ResponseLogger
from programy.utils.console.console import outputLog


class BotClient(ResponseLogger):

    def __init__(self, botid, argument_parser=None):
        self._id = botid
        self._license_keys = LicenseKeys()
        self._storage = None
        self._scheduler = None
        self._email = None
        self._trigger_mgr = None
        self._configuration = None
        self._ping_responder = None

        self._questions = 0

        self._arguments = self.parse_arguments(argument_parser=argument_parser)

        self.initiate_logging(self.arguments)

        self._subsitutions = Substitutions()
        if self.arguments.substitutions is not None:
            self._subsitutions.load_substitutions(self.arguments.substitutions)

        self.load_configuration(self.arguments)
        self.parse_configuration()

        self.load_storage()

        self._bot_factory = BotFactory(self, self.configuration.client_configuration)

        self.load_license_keys()
        self.get_license_keys()
        self._configuration.client_configuration.check_for_license_keys(self._license_keys)

        self.load_scheduler()

        self._renderer = None
        self.load_renderer()

        self.load_email()

        self.load_trigger_manager()

        self.load_ping_responder()

    @staticmethod
    def ylogger_type():
        return "client"

    @property
    def num_questions(self):
        return self._questions

    @property
    def configuration(self):
        return self._configuration

    @property
    def id(self):
        return self._id

    @property
    def arguments(self):
        return self._arguments

    @property
    def license_keys(self):
        return self._license_keys

    @property
    def scheduler(self):
        return self._scheduler

    @property
    def storage_factory(self):
        return self._storage

    @property
    def bot_factory(self):
        return self._bot_factory

    @property
    def renderer(self):
        return self._renderer

    @property
    def trigger_manager(self):
        return self._trigger_mgr

    @property
    def ping_responder(self):
        return self._ping_responder

    def get_description(self):
        if self.configuration is not None:
            if self.configuration.client_configuration is not None:
                return self.configuration.client_configuration.description
        return "Bot Client"

    def get_question_counts(self):
        return self.bot_factory.get_question_counts()

    def add_client_arguments(self, parser=None):
        del parser
        # Nothing to add
        return

    def parse_configuration(self):
        # Nothing to add
        return

    def parse_args(self, arguments, parsed_args):
        del arguments
        del parsed_args
        # Nothing to add
        return

    def parse_arguments(self, argument_parser):
        client_args = CommandLineClientArguments(self, parser=argument_parser)
        client_args.parse_args(self)
        return client_args

    def load_license_keys(self):
        if self.storage_factory.entity_storage_engine_available(StorageFactory.LICENSE_KEYS) is True:
            storage_engine = self.storage_factory.entity_storage_engine(StorageFactory.LICENSE_KEYS)
            if storage_engine is not None:
                keys_store = storage_engine.license_store()
                keys_store.load(self._license_keys)
            else:
                YLogger.error(self, "No storage engine available for license_keys")
        else:
            YLogger.warning(self, "No storage factory setting for license_keys")

    def get_license_keys(self):
        return

    def initiate_logging(self, arguments):
        if arguments.logging is not None:
            try:
                with open(arguments.logging, 'r+', encoding="utf-8") as yml_data_file:
                    logging_config = yaml.load(yml_data_file, Loader=yaml.FullLoader)
                    logging.config.dictConfig(logging_config)
                    YLogger.info(self, "Now logging under configuration")

            except Exception as excep:
                YLogger.exception(self, "Failed to open logging configuration [%s]", excep, arguments.logging)

        else:
            outputLog(self, "Warning. No logging configuration file defined, using defaults...")

    def get_client_configuration(self):
        """
        By overriding this class in you Configuration file, you can add new configurations
        and stil use the dynamic loader capabilities
        :return: Client configuration object
        """
        raise NotImplementedError("You must override this and return a subclassed client configuration")  # pragma: no cover

    def load_configuration(self, arguments, subs: Substitutions = None):
        if arguments.bot_root is None:
            if arguments.config_filename is not None:
                arguments.bot_root = os.path.dirname(arguments.config_filename)
            else:
                arguments.bot_root = "."
            outputLog(self, "No bot root argument set, defaulting to [%s]" % arguments.bot_root)

        if arguments.config_filename is not None:
            self._configuration = ConfigurationFactory.load_configuration_from_file(self.get_client_configuration(),
                                                                                    arguments.config_filename,
                                                                                    arguments.config_format,
                                                                                    arguments.bot_root,
                                                                                    subs)
        else:
            outputLog(self, "No configuration file specified, using defaults only !")
            self._configuration = ProgramyConfiguration(self.get_client_configuration())

    def load_scheduler(self):
        if self.configuration.client_configuration.scheduler is not None:
            YLogger.debug(None, "Loading Scheduler")
            self._scheduler = ProgramyScheduler(self, self.configuration.client_configuration.scheduler)
            self._scheduler.start()

    def load_email(self):
        if self._configuration.client_configuration.email is not None:
            YLogger.debug(None, "Loading Email Manager")
            self._email = EmailSender(self._configuration.client_configuration.email)

    def load_trigger_manager(self):
        if self._configuration.client_configuration.triggers is not None:
            YLogger.debug(None, "Loading Trigger Manager")
            self._trigger_mgr = TriggerManager.load_trigger_manager(self._configuration.client_configuration.triggers)

    def load_storage(self):
        self._storage = StorageFactory()
        if self.configuration.client_configuration.storage is not None:
            YLogger.debug(None, "Loading Storage Factory")
            self._storage.load_engines_from_config(self.configuration.client_configuration.storage)
        else:
            outputLog(self, "No storage defined!")

    def _render_callback(self):
        return False

    def load_renderer(self):
        callback = self._render_callback()
        try:
            if self._configuration.client_configuration.renderer is not None:
                YLogger.debug(None, "Loading Renderer")
                clazz = ClassLoader.instantiate_class(self._configuration.client_configuration.renderer)
                if callback is True:
                    self._renderer = clazz(self)
                else:
                    self._renderer = clazz(None)
                return

        except Exception as e:
            YLogger.exception(None, "Failed to load config specified renderer", e)

        self._renderer = self.get_default_renderer(callback)

    def get_default_renderer(self, callback=True):
        if callback is True:
            return TextRenderer(self)
        else:
            return TextRenderer(None)

    def create_client_context(self, userid):
        client_context = ClientContext(self, userid)
        client_context.bot = self._bot_factory.select_bot()
        client_context.brain = client_context.bot.brain_factory.select_brain()
        return client_context

    def load_ping_responder(self):

        self._ping_responder = PingResponder(self)

        self._ping_responder.register_with_healthchecker()

    def startup(self):
        if self._trigger_mgr is not None:
            self._trigger_mgr.trigger(event=SystemTriggers.SYSTEM_STARTUP)

        if self._ping_responder is not None:
            PingResponder.init_ping_response(self._ping_responder)

    def shutdown(self):
        if self._trigger_mgr is not None:
            self._trigger_mgr.trigger(event=SystemTriggers.SYSTEM_SHUTDOWN)

        if self._ping_responder is not None:
            self._ping_responder.shutdown_ping_service()

    def run(self, app=None):
        del app
        raise NotImplementedError("You must override this and implement the logic to run the client")  # pragma: no cover



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions
of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from typing import Any
from bs4 import BeautifulSoup as Soup
from bs4.element import Tag, NavigableString
from programy.context import ClientContext
from programy.utils.console.console import outputLog


class RichMediaRenderer:

    def __init__(self, callback):
        self._client = callback

    def _default_output(self) -> str:
        return ""

    def _concat_result(self, first, second) -> str:
        del first
        del second
        return ""

    def render(self, client_context, message):
        if message:
            message = "<content>%s</content>"%message
            soup = Soup(message, "lxml-xml")
            if soup.children:
                output = self._default_output()

                for outer_child in soup.children:
                    if outer_child.children:
                        for child in outer_child.children:
                            if isinstance(child, Tag):
                                result = self.parse_tag(client_context, child)
                                if result is not None:
                                    output = self._concat_result(output, result)

                            elif isinstance(child, NavigableString):
                                result = self.parse_text(client_context, child)
                                if result is not None:
                                    output = self._concat_result(output, result)

                return output

            return self.parse_text(client_context, message)

        return None

    def parse_tag(self, client_context: ClientContext, tag) -> Any:

        if tag.name == 'button':
            return self.parse_button(client_context, tag)

        elif tag.name == 'link':
            return self.parse_link(client_context, tag)

        elif tag.name == 'image':
            return self.parse_image(client_context, tag)

        elif tag.name == 'video':
            return self.parse_video(client_context, tag)

        elif tag.name == 'card':
            return self.parse_card(client_context, tag)

        elif tag.name == 'carousel':
            return self.parse_carousel(client_context, tag)

        elif tag.name == 'reply':
            return self.parse_reply(client_context, tag)

        elif tag.name == 'delay':
            return self.parse_delay(client_context, tag)

        elif tag.name == 'split':
            return self.parse_split(client_context, tag)

        elif tag.name == 'list':
            return self.parse_list(client_context, tag)

        elif tag.name == 'olist':
            return self.parse_olist(client_context, tag)

        elif tag.name == 'location':
            return self.parse_location(client_context, tag)

        elif tag.name == 'tts':
            return self.parse_tts(client_context, tag)

        return self.parse_xml(client_context, tag)

    def parse_text(self, client_context: ClientContext, text) -> Any:
        return self.handle_text(client_context, {"type": "text", "text": text})

    def parse_xml(self, client_context, tag):
        text = str(tag)
        return self.handle_text(client_context, {"type": "text", "text": text})

    def extract_button_info(self, tag):
        text = None
        url = None
        postback = None
        for child in tag.children:

            if child.name is None:
                pass

            elif child.name == 'text':
                text = child.text

            elif child.name == 'url':
                url = child.text

            elif child.name == 'postback':
                postback = child.text

            else:
                outputLog(self, "Unknown button tag %s" % child.name)

        data = {"type": "button", "text": text, "url": url, "postback": postback}
        self.extract_class_attr(tag, data)
        self.extract_id_attr(tag, data)
        return data

    def extract_class_attr(self, tag, data):
        if 'class' in tag.attrs:
            data['class'] = tag.attrs['class']

    def extract_id_attr(self, tag, data):
        if 'id' in tag.attrs:
            data['id'] = tag.attrs['id']

    def parse_button(self, client_context, tag):
        button = self.extract_button_info(tag)
        if button['url'] is not None:
            return self.handle_url_button(client_context, button)
        else:
            return self.handle_postback_button(client_context, button)

    def extract_link_info(self, tag):
        text = None
        url = None
        for child in tag.children:

            if child.name is None:
                pass

            elif child.name == 'text':
                text = child.text

            elif child.name == 'url':
                url = child.text

            else:
                outputLog(self, "Unknown link tag %s" % child.name)

        data = {"type": "link", "text": text, "url": url}
        self.extract_class_attr(tag, data)
        self.extract_id_attr(tag, data)
        return data

    def parse_link(self, client_context, tag):
        link = self.extract_link_info(tag)
        return self.handle_link(client_context, link)

    def extract_image_info(self, tag):
        data = {"type": "image", "url": tag.text.strip()}
        self.extract_class_attr(tag, data)
        self.extract_id_attr(tag, data)
        return data

    def parse_image(self, client_context, tag):
        image = self.extract_image_info(tag)
        return self.handle_image(client_context, image)

    def extract_video_info(self, tag):
        data = {"type": "video", "url": tag.text.strip()}
        self.extract_class_attr(tag, data)
        self.extract_id_attr(tag, data)
        return data

    def parse_video(self, client_context, tag):
        video = self.extract_video_info(tag)
        return self.handle_video(client_context, video)

    def extract_card_info(self, tag):
        image = None
        title = None
        subtitle = None
        buttons = []

        for child in tag.children:

            if child.name is None:
                pass

            elif child.name == 'image':
                image = child.text

            elif child.name == 'title':
                title = child.text

            elif child.name == 'subtitle':
                subtitle = child.text

            elif child.name == 'button':
                button = self.extract_button_info(child)
                buttons.append(button)

            else:
                outputLog(self, "Unknown card tag [%s]" % child.name)

        data = {"type": "card", "image": image, "title": title, "subtitle": subtitle, "buttons": buttons}
        self.extract_class_attr(tag, data)
        self.extract_id_attr(tag, data)
        return data

    def parse_card(self, client_context, tag):
        card = self.extract_card_info(tag)
        return self.handle_card(client_context, card)

    def extract_carousel_info(self, tag):
        cards = []
        for child in tag.children:

            if child.name is None:
                pass

            elif child.name == 'card':
                card = self.extract_card_info(child)
                cards.append(card)

            else:
                outputLog(self, "Unknown carousel tag %s" % child.name)

        data = {"type": "carousel", "cards": cards}
        self.extract_class_attr(tag, data)
        self.extract_id_attr(tag, data)
        return data

    def parse_carousel(self, client_context, tag):
        carousel = self.extract_carousel_info(tag)
        return self.handle_carousel(client_context, carousel)

    def extract_reply_info(self, tag):
        text = None
        postback = None
        for child in tag.children:

            if child.name is None:
                pass

            elif child.name == 'text':
                text = child.text.strip()

            elif child.name == 'postback':
                postback = child.text.strip()

            else:
                outputLog(self, "Unknown reply tag %s" % child.name)

        data = {"type": "reply", "text": text, "postback": postback}
        self.extract_class_attr(tag, data)
        self.extract_id_attr(tag, data)
        return data

    def parse_reply(self, client_context, tag):
        reply = self.extract_reply_info(tag)
        return self.handle_reply(client_context, reply)

    def extract_delay_info(self, tag):
        seconds = None
        for child in tag.children:

            if child.name is None:
                pass

            elif child.name == 'seconds':
                seconds = child.text.strip()

        data = {"type": "delay", "seconds": seconds}
        self.extract_class_attr(tag, data)
        self.extract_id_attr(tag, data)
        return data

    def parse_delay(self, client_context, tag):
        delay = self.extract_delay_info(tag)
        return self.handle_delay(client_context, delay)

    def parse_split(self, client_context, tag):
        split = {"type": "split"}
        self.extract_class_attr(tag, split)
        self.extract_id_attr(tag, split)
        return self.handle_split(client_context, split)

    def extract_item_info(self, tag):

        if tag.name == 'button':
            return self.extract_reply_info(tag)

        elif tag.name == 'link':
            return self.extract_link_info(tag)

        elif tag.name == 'image':
            data = {"type": "image", "url": tag.text}
            self.extract_class_attr(tag, data)
            self.extract_id_attr(tag, data)
            return data

        elif tag.name == 'video':
            data =  {"type": "video", "url": tag.text}
            self.extract_class_attr(tag, data)
            self.extract_id_attr(tag, data)
            return data

        elif tag.name == 'card':
            return self.extract_card_info(tag)

        elif tag.name == 'carousel':
            return self.extract_carousel_info(tag)

        elif tag.name == 'reply':
            return self.extract_reply_info(tag)

        elif tag.name == 'delay':
            return self.extract_delay_info(tag)

        elif tag.name == 'split':
            # Not allowed
            pass

        elif tag.name == 'list':
            return self.extract_list_info(tag)

        elif tag.name == 'olist':
            return self.extract_list_info(tag)

        elif tag.name == 'location':
            # Not allowed
            pass

        else:
            if isinstance(tag, Tag):
                text = tag.text

            elif isinstance(tag, NavigableString):
                text = tag

            data = {"type": "text", "text": text}
            self.extract_class_attr(tag, data)
            self.extract_id_attr(tag, data)
            return data

    def extract_list_info(self, tag):
        items = []

        for child in tag.children:

            if child.name is None:
                pass

            elif child.name == 'item':

                for childs_child in child.children:

                    if isinstance(childs_child, Tag):
                        items.append(self.extract_item_info(childs_child))

                    elif isinstance(childs_child, NavigableString):
                        childs_child_text = childs_child.strip()
                        if childs_child_text:
                            items.append({'type': 'text', 'text': childs_child_text})

            else:
                outputLog(self, "Unknown list tag %s" % child.name)

        data = {'type': 'list', 'items': items}
        self.extract_class_attr(tag, data)
        self.extract_id_attr(tag, data)
        return data

    def parse_list(self, client_context, tag):
        rendered = self.extract_list_info(tag)
        return self.handle_list(client_context, rendered)

    def parse_olist(self, client_context, tag):
        rendered = self.extract_list_info(tag)
        return self.handle_ordered_list(client_context, rendered)

    def parse_location(self, client_context, tag):
        del tag
        return self.handle_location(client_context, {"type": "location"})

    def parse_tts(self, client_context, tag):
        return self.handle_tts(client_context, {"type": "tts", "text": tag.text})

    ######################################################################################################
    # You need to implement all of these and decide how to display the various rich media elements
    # 
    def handle_text(self, client_context, text):
        del client_context
        del text
        return None

    def handle_url_button(self, client_context, button):
        del client_context
        del button
        return None

    def handle_postback_button(self, client_context, button):
        del client_context
        del button
        return None

    def handle_link(self, client_context, link):
        del client_context
        del link
        return None

    def handle_image(self, client_context, image):
        del client_context
        del image
        return None

    def handle_video(self, client_context, video):
        del client_context
        del video
        return None

    def handle_card(self, client_context, card):
        del client_context
        del card
        return None

    def handle_carousel(self, client_context, carousel):
        del client_context
        del carousel
        return None

    def handle_reply(self, client_context, reply):
        del client_context
        del reply
        return None

    def handle_delay(self, client_context, delay):
        del client_context
        del delay
        return None

    def handle_split(self, client_context, split):
        del client_context
        del split
        return None

    def handle_list(self, client_context, lst):
        del client_context
        del lst
        return None

    def handle_ordered_list(self, client_context, lst):
        del client_context
        del lst
        return None

    def handle_location(self, client_context, location):
        del client_context
        del location
        return None

    def handle_tts(self, client_context, tts):
        del client_context
        del tts
        return None
